﻿//Copyright (c) 2009, Mark S. Burrell
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are 
//permitted provided that the following conditions are met:

//    * Redistributions of source code must retain the above copyright notice, this list 
//of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this 
//list of conditions and the following disclaimer in the documentation and/or other 
//materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
//EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
//OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
//SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
//OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Sedici.Hardware;


namespace Assembler
{
    class Assembler
    {

        static SediciAssembler assembler = new SediciAssembler();

        static void Main(string[] args)
        {

            // variables for argument strings, plus a string separator
            string[] rawarguments = { };
            string[] arguments = { };
            string[] separator = { " ", "\t" };
            string[] equals = { "=" };

            bool machineCodeListing = false;

            string assemblyfile = "";

            string message = "";

            // check if we have a file as the first param
            if (args.Length > 0)
            {
                args[0] = convertSynonyms(args[0]);
                switch (args[0])
                {
                    case "/help":
                        showHelp();
                        break;
                    default:
                        assemblyfile = args[0];
                        break;
                }

            }
            else
            {
                // no params so show help by default
                showHelp();
            }

            // now look for extra params
            if (args.Length > 1)
            {
                //Check parameters
                for (int i = 1; i < args.Length; i++)
                {
                    // if the argument includes a '=' then separate it out first
                    arguments = args[i].Split(equals, StringSplitOptions.None);
                    arguments[0] = convertSynonyms(arguments[0]);

                    switch (arguments[0])
                    {
                        case "/mc":
                            machineCodeListing = true;
                            break;
                        default:
                            message = "Unknown parameter [" + args[i] + "]";
                            Console.Error.WriteLine(message);
                            break;
                    }
                }
            }
            else
            {
                // no further params
            }

            if (assemblyfile == "") {
                //Console.WriteLine("No file to assemble");
            }
            else {

                supportMethods.updateFileLocation(ref assemblyfile, globalConstants.envName);

                if (File.Exists(assemblyfile))
                {
                    if (machineCodeListing)
                    {
                        assembler.setOutputToMachineCode(true);
                    }

                    string firmware = "firmware.frm";


                    supportMethods.updateFileLocation(ref firmware, globalConstants.envName);
                    assembler.loadFirmware(firmware);

                    assembler.loadAssemblyCode(assemblyfile);

                    string output = assembler.assemble();

                    if (assembler.Errors == 0)
                    {
                        Console.Error.WriteLine("Assembled with no errors");
                    }
                    else if (assembler.Errors == 1)
                    {
                        Console.Error.WriteLine("Assembled with 1 error");
                    }
                    else
                    {
                        Console.Error.WriteLine("Assembled with " + assembler.Errors + " errors");
                    }



                    Console.WriteLine(output);
                }
                else
                {
                    message = "Source file does not exist";
                    Console.Error.WriteLine(message);
                }
            }

            // end main
        }

        static void showHelp()
        {

            Console.WriteLine();
            Console.WriteLine(" NAME");
            Console.WriteLine();
            Console.WriteLine("   " + assembler.Name);
            Console.WriteLine();
            Console.WriteLine(" DESCRIPTION");
            Console.WriteLine("   This assembler writes output that can be loaded and");
            Console.WriteLine("   executed by a Sedici processor based system.");
            Console.WriteLine();
            Console.WriteLine(" USAGE");
            Console.WriteLine();
            Console.WriteLine("   sasm [ (srcfile | /h*elp | /? ) ] [/mc]");
            Console.WriteLine();
            Console.WriteLine("        srcfile : the assembly file to be assembled");
            Console.WriteLine("        /help   : list this help information");
            Console.WriteLine("        /mc     : output is machine code only");
            Console.WriteLine();
            Console.WriteLine("   For example, if you had a file called a.src to");
            Console.WriteLine("   assemble you would type:");
            Console.WriteLine();
            Console.WriteLine("       sasm a.src");
            Console.WriteLine();
            Console.WriteLine("   To capture the output to a separate file use:");
            Console.WriteLine();
            Console.WriteLine("       sasm a.src > a.asm");
            Console.WriteLine();
            Console.WriteLine("   The resulting file, if there are no assembly errors,");
            Console.WriteLine("   can be executed in a Sedici based system, such as");
            Console.WriteLine("   the Sedici console program.");
            Console.WriteLine();
            Console.WriteLine("   Running under Mono you simply replace 'sasm' in the");
            Console.WriteLine("   examples above with 'mono sasm.exe'");
            Console.WriteLine();
            Console.WriteLine("   It is assumed that the firmware file loads successfully.");
            Console.WriteLine("   This is free software and comes with no warranty.");
            Console.WriteLine();
            Console.WriteLine(" COPYRIGHT");
            Console.WriteLine();
            Console.WriteLine("   " + globalConstants.COPYRIGHT);
            Console.WriteLine();
            Console.WriteLine(" VERSION");
            Console.WriteLine();
            Console.WriteLine("   " + globalConstants.VERSION);
            Console.WriteLine();

            return;
        }

        static string convertSynonyms(string cmd)
        {
            string returnValue = cmd;
            switch (cmd)
            {
                case "/?":
                case "/h":
                case "/he":
                case "/hel":
                case "/help":
                    returnValue = "/help";
                    break;
                default:
                    break;
            }
            return (returnValue);
        }



    }
}
