// The Sedici console

//Copyright (c) 2009, Mark S. Burrell
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are 
//permitted provided that the following conditions are met:

//    * Redistributions of source code must retain the above copyright notice, this list 
//of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this 
//list of conditions and the following disclaimer in the documentation and/or other 
//materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
//EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
//OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
//SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
//OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.Collections;
using System.IO;

using Sedici.Hardware;

namespace SediciConsole
{

    public class TextConsole
    {
        const string systemName = "Sedici Console";
        const string programName = "console";

        static string HCOfile = "console.tlp";
        static string instructionSet = "firmware.frm";

        static ulong Cycles = 0;

        static Messages messageList = new Messages();

        // instantiate Memory and CPU

        static Memory     Memory = new Memory();
        static CPU CPU = new CPU(ref messageList);

        // Methods
        public static void Main(string[] args)
        {
            #region Local variables

            // initialisation
            bool programFinished = false;

            // variables for argument strings, plus a string separator
            string rawInputString = "";
            string[] rawarguments = { };
            string[] arguments = { };
            string[] separator = { " ", "\t" };
            string[] equals = { "=" };

            bool paramGo = false;
            bool paramQuit = false;
            bool paramNodefault = false;

            string message = "";
            #endregion

            #region Install the memory-mapped chips

            // all installed by default, so turn off those not wanted
            Memory.graphicsChip.Installed = false;

            // perfrom a hard reset to power everything up cleanly
            HardReset();

            #endregion

            // ensure that the micromemory is empty
            CPU.Wipe();
            showMessages();

            // display the console version information
            cmdVer();

            # region Check for help and nodefault parameters
            //check help and nodefault parameters here
            for (int i = 0; i < args.Length; i++)
            {
                args[i] = convertSynonyms(args[i]);
                switch (args[i])
                {
                    case "/help":
                        Help();
                        break;
                    case "/nodefault":
                        paramNodefault = true;
                        break;
                    default:
                        break;
                }
            }
            #endregion

            #region Load the default instruction set if required
            if (paramNodefault)
            {
                Console.WriteLine("Default instruction set not loaded");
            }
            else
            {
                // load the instruction set if the file exists
                supportMethods.updateFileLocation(ref instructionSet, globalConstants.envName);
                CPU.LoadFirmware(instructionSet);
                showMessages();

            }
            #endregion

            #region Check for other parameters
            // parameter run requested
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument includes a '=' then separate it out first
                arguments = args[i].Split(equals, StringSplitOptions.RemoveEmptyEntries);
                arguments[0] = convertSynonyms(arguments[0]);

                switch (arguments[0])
                {
                    case "/help":
                        // already dealt with
                        break;
                    case "/nodefault":
                        // already dealt with
                        break;
                    case "/pc":
                        if (arguments.Length > 1)
                        {
                            cmdRegister(CPURegister.PC, arguments);
                            Console.WriteLine("PC=" + CPU.GetRegister(CPURegister.PC, Base.hex));
                            showMessages();

                        }
                        else
                        {
                            Console.WriteLine("PC update argument missing");
                        }
                        break;
                    case "/load":
                        if (arguments.Length > 1)
                        {
                            supportMethods.updateFileLocation(ref arguments[1], globalConstants.envName);
                            LoadProgram(arguments[1], ref CPU);

                        }
                        break;
                    case "/frm":
                        if (arguments.Length > 1)
                        {
                            supportMethods.updateFileLocation(ref arguments[1], globalConstants.envName);
                            CPU.LoadFirmware(arguments[1]);
                            showMessages();

                        }
                        break;
                    case "/go":
                        paramGo = true;
                        break;
                    case "/quit":
                        paramQuit = true;
                        break;
                    default:
                        message = "Unknown parameter [" + args[i] + "]";
                        Console.WriteLine(message);
                        break;
                }
            }
            #endregion

            #region Automate the run process if required
            // run the loaded program if paramGo is set
            if (paramGo) {
                Console.WriteLine("Go requested...");
                if (paramQuit)
                {
                    Console.WriteLine("Quit after halt requested...");
                }

                // run the processor
                cmdRun();
            }

            if (paramQuit)
            {
                Environment.Exit(0);
            }

            #endregion

            // display a prompt
            Console.WriteLine();
            Console.Write(">");

            #region The main program loop

            // begin the command loop
            while (!programFinished) {

                // get the command, store it (raw as well as lowercase), then parse it
                rawInputString = Console.ReadLine();
                rawarguments   = rawInputString.Split(separator, StringSplitOptions.RemoveEmptyEntries);
                rawInputString = rawInputString.ToLower();
                arguments      = rawInputString.Split(separator, StringSplitOptions.RemoveEmptyEntries);

                // if we have a command, then see if it's valid and do something about it
                if (arguments.Length > 0) {
                    arguments[0] = convertSynonyms(arguments[0]);
                    switch (arguments[0]) {
                        case "help":
                            cmdHelp(arguments);
                            break;
                        case "comment":
                            Console.WriteLine("[" + CPU.firmware.Comment + "]");
                            break;
                        case "version":
                            cmdVer();
                            break;
                        case "df":
                            cmdDF(arguments);
                            break;
                        case "jasm":
                            cmdJasm(arguments);
                            break;
                        case "jcc":
                            cmdJCC(arguments);
                            break;
                        case "md":
                            cmdMD(arguments);
                            break;
                        case ".d0":
                            cmdRegister(CPURegister.D0, arguments);
                            break;
                        case ".d1":
                            cmdRegister(CPURegister.D1, arguments);
                            break;
                        case ".d2":
                            cmdRegister(CPURegister.D2, arguments);
                            break;
                        case ".d3":
                            cmdRegister(CPURegister.D3, arguments);
                            break;
                        case ".d4":
                            cmdRegister(CPURegister.D4, arguments);
                            break;
                        case ".d5":
                            cmdRegister(CPURegister.D5, arguments);
                            break;
                        case ".d6":
                            cmdRegister(CPURegister.D6, arguments);
                            break;
                        case ".d7":
                            cmdRegister(CPURegister.D7, arguments);
                            break;
                        case ".mdr":
                            cmdRegister(CPURegister.MDR, arguments);
                            break;
                        case ".mar":
                            cmdRegister(CPURegister.MAR, arguments);
                            break;
                        case ".inc":
                            cmdRegister(CPURegister.INC, arguments);
                            break;
                        case ".pc":
                            cmdRegister(CPURegister.PC, arguments);
                            break;
                        case ".alux":
                            cmdRegister(CPURegister.ALUx, arguments);
                            break;
                        case ".aluy":
                            cmdRegister(CPURegister.ALUy, arguments);
                            break;
                        case ".alu":
                            cmdRegister(CPURegister.ALUr, arguments);
                            break;
                        case ".usp":
                            cmdRegister(CPURegister.USP, arguments);
                            break;
                        case ".ssp":
                            cmdRegister(CPURegister.SSP, arguments);
                            break;
                        case ".cu":
                            cmdRegister(CPURegister.CU, arguments);
                            break;
                        case ".ircmd":
                            cmdRegister(CPURegister.IRcmd, arguments);
                            break;
                        case ".iroperand":
                            cmdRegister(CPURegister.IRoperand, arguments);
                            break;
                        case ".psr":
                            cmdRegister(CPURegister.PSR, arguments);
                            break;
                        case "reset":
                            cmdReset();
                            break;
                        case ".n":
                            cmdFlag(PSRflag.N, arguments);
                            break;
                        case ".v":
                            cmdFlag(PSRflag.V, arguments);
                            break;
                        case ".z":
                            cmdFlag(PSRflag.Z, arguments);
                            break;
                        case ".c":
                            cmdFlag(PSRflag.C, arguments);
                            break;
                        case ".i":
                            cmdFlag(PSRflag.I, arguments);
                            break;
                        case ".e":
                            cmdFlag(PSRflag.E, arguments);
                            break;
                        // temporarily removed until useful supervisor mode
                        //case ".s":
                        //    cmdFlag(PSRflag.S, arguments);
                        //    break;
                        case "frm":
                            cmdMCO(arguments);
                            break;
                        case "frm2":
                            cmdMCO2(arguments);
                            break;
                        case "load":
                            cmdLoad(arguments);
                            break;
                        case "run":
                            cmdRun();
                            break;
                        case "trace":
                            cmdTrace();
                            break;
                        case "vmic":
                            cmdVmic(arguments);
                            break;
                        case "vmic2":
                            cmdVmicRaw(arguments);
                            break;
                        case "bm":
                            cmdBM(arguments);
                            break;
                        case "bf":
                            cmdBF(arguments);
                            break;
                        case "cd":
                            cmdCD(arguments);
                            break;
                        case "ls":
                            cmdLS();
                            break;
                        case "envls":
                            cmdEnvLS();
                            break;
                        case "view":
                            cmdView(arguments);
                            break;
                        case "about":
                            cmdAbout(arguments);
                            break;
                        case "code":
                            cmdCode(arguments, rawarguments);
                            break;
                        case "fetch":
                            cmdFetch();
                            break;
                        case "execute":
                            cmdExecute();
                            break;
                        case "jumpers":
                        case "memupdate":
                        case "periph":
                        case "ioboard":
                            cmdObsolete();
                            break;
                        case "vec":
                            cmdVec(arguments);
                            break;
                        case "quit":
                            programFinished = true;
                            break;
                        case "clrmm":
                            cmdClrmm();
                            break;
                        case "power":
                            HardReset();
                            Console.WriteLine(Environment.NewLine + "System hard reset");
                            break;
                        case "devices":
                            cmdDevices();
                            break;
                        case "toggle":
                            cmdToggle(arguments);
                            break;
                        case "base":
                            cmdBase(arguments);
                            break;
                        case "intbase":
                            cmdIntBase(arguments);
                            break;
                        case "cls":
                            Console.Clear();
                            break;
                        default:
                            // don't know what the command is
                            message = "Unknown command \"" + arguments[0] + "\"";
                            Console.WriteLine(message);
                            break;
                    }
                }

                // write a new prompt
                Console.WriteLine();
                Console.Write(">");
            }
            // end of the command loop
            #endregion

        } // end Main

        static void cmdAbout(string[] arguments)
        {
            int i;
            int size = arguments.Length + 1;
            string[] args = new string[size];

            for (i = 0; i < arguments.Length; i++)
            {
                args[i] = arguments[i];
            }
            args[i] = "about";
            cmdHelp(args);
            return;
        }

        static void cmdFetch()
        {
            CPU.RunFetchCycle(ref Memory);
            showMessages();

            cmdDF();
            return;
        }

        static void cmdExecute()
        {
            CPU.RunExecuteCycle(ref Memory);
            showMessages();

            cmdDF();
            return;
        }

        static void cmdObsolete()
        {
            Console.WriteLine("This command is obsolete");
            return;
        }

        static void cmdCD()
        {
            // path for accessing current working directory
            string path = "";

            path = Directory.GetCurrentDirectory();
            Console.WriteLine("Current working directory: " + path);
            return;
        }

        static void cmdClrmm()
        {
            CPU.Wipe();
            showMessages();

            Console.WriteLine("Micromemory contents erased");
            return;
        }

        static void cmdDevices()
        {
            int baseAddress = 0;
            int endAddress = 0;
            int length = 0;

            // display the bases
            Console.WriteLine();
            Console.WriteLine("Device addresses          start| end  | len  | status");
            Console.WriteLine("------------------------------------------------------------");

            Console.Write("I/O chip (io)             ");
            baseAddress = Memory.ioChip.BaseAddress;
            length = Memory.ioChip.Size;
            endAddress = baseAddress + length - 1;
            Console.Write(supportMethods.FormatUshort((ushort)baseAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)endAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)length, Base.hex));

            Console.Write(" | ");
            if (Memory.ioChip.Installed == true)
            {
                Console.WriteLine("installed");
            }
            else
            {
                Console.WriteLine("not installed");
            }


            Console.Write("Clock chip (clock)        ");
            baseAddress = Memory.clockChip.BaseAddress;
            length = Memory.clockChip.Size;
            endAddress = baseAddress + length - 1;
            Console.Write(supportMethods.FormatUshort((ushort)baseAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)endAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)length, Base.hex));
            Console.Write(" | ");
            if (Memory.clockChip.Installed == true)
            {
                Console.WriteLine("installed");
            }
            else
            {
                Console.WriteLine("not installed");
            }

            Console.Write("ROM chip (rom)            ");
            baseAddress = Memory.romChip.BaseAddress;
            length = Memory.romChip.Size;
            endAddress = baseAddress + length - 1;
            Console.Write(supportMethods.FormatUshort((ushort)baseAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)endAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)length, Base.hex));
            Console.Write(" | ");
            if (Memory.romChip.Installed == true)
            {
                Console.WriteLine("installed");
            }
            else
            {
                Console.WriteLine("not installed");
            }

            Console.Write("Graphics chip (graphics)  ");
            baseAddress = Memory.graphicsChip.BaseAddress;
            length = Memory.graphicsChip.Size;
            endAddress = baseAddress + length - 1;
            Console.Write(supportMethods.FormatUshort((ushort)baseAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)endAddress, Base.hex));
            Console.Write(" | ");
            Console.Write(supportMethods.FormatUshort((ushort)length, Base.hex));
            Console.Write(" | ");
            if (Memory.graphicsChip.Installed == true)
            {
                Console.WriteLine("installed");
            }
            else
            {
                Console.WriteLine("not installed");
            }
        }


        static void cmdToggle(string[] arguments) {
            Chip component;

            if (arguments.Length == 2)
            {
                switch (arguments[1])
                {
                    case "io":
                        component = Memory.ioChip;
                        break;
                    case "graphics":
                        component = Memory.graphicsChip;
                        break;
                    case "clock":
                        component = Memory.clockChip;
                        break;
                    case "rom":
                        component = Memory.romChip;
                        break;
                    default:
                        Console.WriteLine("Unknown device.  Should be io, graphics, clock or rom");
                        return;
                }

                if (component.Installed == true)
                {
                    component.Installed = false;
                    Console.WriteLine("Device removed");
                }
                else
                {
                    component.Installed = true;
                    Console.WriteLine("Device installed");
                }
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdBase(string[] arguments)
        {
            Chip component;
            ushort result = 0;

            if ((arguments.Length < 2) || (arguments.Length > 3))
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            else
            {
                switch (arguments[1])
                {
                    case "io":
                        component = Memory.ioChip;
                        break;
                    case "graphics":
                        component = Memory.graphicsChip;
                        break;
                    case "clock":
                        component = Memory.clockChip;
                        break;
                    case "rom":
                        component = Memory.romChip;
                        break;
                    default:
                        Console.WriteLine("Unknown device.  Should be io, graphics, clock or rom");
                        return;
                }


                if (arguments.Length == 2)
                {
                    Console.WriteLine("Base=" + supportMethods.FormatUshort(component.BaseAddress, Base.hex));
                }
                else if (arguments.Length == 3)
                {
                    if (supportMethods.isNumber(arguments[2], ref result) == 0)
                    {
                        Console.WriteLine("Number format invalid");
                    }
                    else
                    {
                        // do the update
                        component.BaseAddress = result;
                        Console.WriteLine("Base=" + supportMethods.FormatUshort(component.BaseAddress, Base.hex));
                        int check = component.BaseAddress + component.Size - 1;
                        if (check > 0xFFFF)
                        {
                            Console.WriteLine("Card installed off edge of memory, some locations inaccessible");
                        }
                    }

                }
            }
            return;
        }

        static void cmdIntBase(string[] arguments)
        {
            ushort result = 0;
            if (arguments.Length == 1)
            {
                Console.WriteLine("Interrupt Base=" + supportMethods.FormatUshort(CPU.InterruptBase, Base.hex));
            }
            else if (arguments.Length == 2)
            {
                if (supportMethods.isNumber(arguments[1], ref result) == 0)
                {
                    Console.WriteLine("Number format invalid");
                }
                else
                {
                    CPU.InterruptBase = result;
                    Console.WriteLine("Interrupt Base=" + supportMethods.FormatUshort(CPU.InterruptBase, Base.hex));
                    int check = CPU.InterruptBase + 7;
                    if (check > 0xFFFF)
                    {
                        Console.WriteLine("Interrupt base installed off edge of memory, some locations inaccessible");
                    }
                }
            }
            else {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdCode(string[] arguments, string[] rawarguments)
        {
            if (arguments.Length == 2)
            {
                CPU.RunSingleMicroInstruction(rawarguments[1], ref Memory);
                showMessages();

                //checkMessageStore(CPU);
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdVec(string[] arguments)
        {
            if (arguments.Length == 1)
            {
                Console.Write("Vec=" + CPU.Vector);
                Console.WriteLine();
            }
            else if (arguments.Length == 2)
            {
                // needs to convert param into a number and load it into vector
                ushort result = 0;

                if (supportMethods.isNumber(arguments[1], ref result) == 0)
                {
                    Console.WriteLine("Number format invalid");
                }
                else
                {
                    CPU.Vector = result;
                    //checkMessageStore(CPU);
                } 
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
        }

        static void cmdFlag(PSRflag flag, string[] arguments)
        {
            ushort result = 0;
            if (arguments.Length == 1)
            {
                Console.WriteLine(flag + "=" + CPU.GetFlag(flag));
                Console.WriteLine();
            }
            else if (arguments.Length == 2)
            {
                if (supportMethods.isNumber(arguments[1], ref result) == 0)
                {
                    Console.WriteLine("Number format invalid");
                }
                else
                {
                    CPU.SetFlag(flag, result);
                    //checkMessageStore(CPU);
                }
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdVmic(string[] arguments)
        {
            // used to store an opcode's microinstructions
            string[] details = { };

            ushort ushorti = 0;


            if (arguments.Length == 2)
            {
                if (supportMethods.isNumber(arguments[1], ref ushorti) == 0)
                {
                    Console.WriteLine("Number format invalid");
                }
                else
                {
                    details = CPU.OpcodeDetails(ushorti);
                    for (int i = 0; i < details.Length; i++)
                    {
                        Console.WriteLine(details[i]);
                    }
                    //checkMessageStore(CPU);
                }
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
        }

        static void cmdVmicRaw(string[] arguments)
        {
            // used to store an opcode's microinstructions
            string[] details = { };

            ushort ushorti = 0;


            if (arguments.Length == 2)
            {
                if (supportMethods.isNumber(arguments[1], ref ushorti) == 0)
                {
                    Console.WriteLine("Number format invalid");
                }
                else
                {
                    details = CPU.OpcodeRawDetails(ushorti);
                    for (int i = 0; i < details.Length; i++)
                    {
                        Console.WriteLine(details[i]);
                    }
                    //checkMessageStore(CPU);
                }
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
        }

        static void cmdVer()
        {
            Console.WriteLine();
            Console.WriteLine(systemName + " " + globalConstants.VERSION);
            Console.WriteLine(globalConstants.COPYRIGHT);
	        //Console.WriteLine("This is free software and comes with no warranty.");
            Console.WriteLine();
            return;
        }

        static void cmdView(string[] arguments)
        {
            bool fileOutput = false;
            string[] separator = { "file=" };
            string[] filenamestring = { };
            string[] details = { };
            StreamWriter sw;
            string filename = "";

            if (arguments.Length == 1)
            {
                filename = "";
            }
            else if (arguments.Length == 2)
            {
                filename = arguments[1];
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
                return;
            }


            filenamestring = filename.Split(separator, StringSplitOptions.None);
            if (filename != "")
            {
                if (filenamestring.Length < 2)
                {
                    Console.WriteLine("Filename parameter malformed");
                    return;
                }
                else
                {
                    filename = filenamestring[1];
                    if (filename == "")
                    {
                        Console.WriteLine("Missing filename parameter");
                        return;
                    }
                    else
                    {
                        fileOutput = true;
                    }
                }
            }

            if (fileOutput == true)
            {
                try
                {
                    sw = File.CreateText(filename);
                }
                catch (IOException)
                {
                    Console.WriteLine("File creation failed for some reason, malformed filename?");
                    return;
                }
                details = CPU.GetInstructionSetDetails();
                for (int i = 0; i < details.Length; i++)
                {
                    sw.WriteLine(details[i]);
                }
                //checkMessageStore(CPU);
                sw.Close();
            }
            else
            {
                details = CPU.GetInstructionSetDetails();
                for (int i = 0; i < details.Length; i++)
                {
                    Console.WriteLine(details[i]);
                }
                //checkMessageStore(CPU);
            }
            return;
        }

        static void cmdReset() {
            CPU.HardReset();
            showMessages();

            Memory.SoftReset();
            return;
        }

        static void cmdTrace()
        {
            // run a single fetch/execute then display a DF
            CPU.RunFetchExecuteCycle(ref Memory);
            showMessages();

            //checkMessageStore(CPU);
            cmdDF();
            return;
        }

        static void cmdLoad(string[] arguments)
        {
            if (arguments.Length == 2)
            {
                supportMethods.updateFileLocation(ref arguments[1], globalConstants.envName);
                LoadProgram(arguments[1], ref CPU);

            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdMCO(string[] arguments)
        {
            if (arguments.Length == 2)
            {
                supportMethods.updateFileLocation(ref arguments[1], globalConstants.envName);
                CPU.LoadFirmware(arguments[1]);
                showMessages();

            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdMCO2(string[] arguments)
        {
            if (arguments.Length == 2)
            {
                supportMethods.updateFileLocation(ref arguments[1], globalConstants.envName);
                CPU.LoadFirmware2(arguments[1]);
                showMessages();

            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdHelp(string[] arguments)
        {
            string contents = "";
            string[] lines = { };
            string[] line = { };
            string[] separator = { Environment.NewLine };
            //string[] spaceseparator = { " ", "\r\n" };
            string[] spaceseparator = { " " };
            bool correctSection = false;
            bool helpFound = false;

            string message = "";

            string file;
            string helpString = "";

            supportMethods.updateFileLocation(ref HCOfile, globalConstants.envName);
            file = HCOfile;

            if (arguments.Length == 1)
            {
                helpString = "sedici";
            }
            else if (arguments.Length == 2)
            {
                arguments[1] = convertSynonyms(arguments[1]);
                helpString = arguments[1];
            }
            else
            {
                Console.WriteLine("Too many params to help");
                return;
            }


            if (File.Exists(HCOfile))
            {
                helpString = "^" + helpString;

                // read the file
                contents = File.ReadAllText(HCOfile);

                lines = contents.Split(separator, StringSplitOptions.None);

                for (int count = 0; count < lines.Length; count++)
                {

                    line = lines[count].Split(spaceseparator, StringSplitOptions.RemoveEmptyEntries);

                    if (line.Length > 0)
                    {
                        if (line[0].Equals("*", StringComparison.CurrentCultureIgnoreCase))
                        {
                            correctSection = false;
                        }
                        else if (line[0].Equals(helpString, StringComparison.CurrentCultureIgnoreCase))
                        {
                            correctSection = true;
                            helpFound = true;
                        }
                        else if (line[0].Substring(0, 1) == "^")
                        {
                        }
                        else
                        {
                            if (correctSection == true) Console.WriteLine(lines[count]);
                        }
                    }
                    else
                    {
                        if (correctSection == true) Console.WriteLine();
                    }

                }
                if (helpFound == false)
                {
                    Console.WriteLine("No help on that subject");
                    Console.WriteLine();
                }
            }
            else
            {
                message= "Unable to open the help file [" + HCOfile + "]";
                Console.WriteLine(message);
            }

            return;
        }

        static void cmdRun()
        {
            // definition for keyboard events
            ConsoleKeyInfo cki = new ConsoleKeyInfo();

            bool systemReset = false;

            string outputbuffer = "";

            CPU.Start();
            showMessages();

            while(CPU.IsRunning) {
                CPU.RunFetchExecuteCycle(ref Memory);

                // check for keypress
                if (Console.KeyAvailable == true)
                {
                    cki = Console.ReadKey(true); // grab keypress

                    if (cki.Key == ConsoleKey.F2)
                    {
                        // perform a hard reset
                        systemReset = true;
                        HardReset();
                    }
                    else if (cki.Key == ConsoleKey.Escape)
                    {
                        // drop into the built-in monitor
                        CPU.SoftReset();
                    }
                    else if (cki.Key == ConsoleKey.F1)
                    {
                        // raise an NMI
                        CPU.SetFlag(PSRflag.I, 1);
                        CPU.Vector = 7;
                    }
                    else
                    {
                        // put keypress into IDR if card installed
                        if (Memory.ioChip.Installed) {
                            Memory.Write((ushort)(Memory.ioChip.BaseAddress + Chip_IO.IDR), 
                                          ( (ushort)(cki.KeyChar & 0x007F)  ));
                        }
                    }
                }

                // deal with I/O outputBuffer
                if (Memory.ioChip.Installed)
                {
                    outputbuffer = Memory.OutputBuffer();
                    if (outputbuffer != "") {
                        Console.Write(outputbuffer);
                    }
                }
            }
            showMessages();

            if (!systemReset) {
                Console.Write(Environment.NewLine + "Processor halted at : " + CPU.GetRegister(CPURegister.PC, Base.hex));
                Cycles = CPU.MachineCycles;
                Console.WriteLine(" (Machine cycles = " + Cycles.ToString() + ")");
                CPU.MachineCycles = 0;
            }
            return;
        }

        static void cmdDF()
        {
            string[] args = new string[1];

            args[0] = "df";
            cmdDF(args);
            return;
        }

        static void old_cmdDF(string[] arguments)
        {
            ushort opcode;

            if (arguments.Length == 1)
            {

                Console.Write(" D0=" + CPU.GetRegister(CPURegister.D0, Base.hex) + " ");
                Console.Write(" D1=" + CPU.GetRegister(CPURegister.D1, Base.hex) + " ");
                Console.Write(" D2=" + CPU.GetRegister(CPURegister.D2, Base.hex) + " ");
                Console.Write("  D3=" + CPU.GetRegister(CPURegister.D3, Base.hex) + " ");
                Console.Write("  D4=" + CPU.GetRegister(CPURegister.D4, Base.hex) + " ");
                Console.Write(" D5=" + CPU.GetRegister(CPURegister.D5, Base.hex) + " ");
                Console.Write(" D6=" + CPU.GetRegister(CPURegister.D6, Base.hex) + " ");
                Console.Write(" D7=" + CPU.GetRegister(CPURegister.D7, Base.hex) + " ");
                Console.WriteLine();

                // next line of registers
                Console.Write(" PC=" + CPU.GetRegister(CPURegister.PC, Base.hex) + " ");
                Console.Write("MAR=" + CPU.GetRegister(CPURegister.MAR, Base.hex) + " ");
                Console.Write("MDR=" + CPU.GetRegister(CPURegister.MDR, Base.hex) + " ");
                Console.Write("ALUx=" + CPU.GetRegister(CPURegister.ALUx, Base.hex) + " ");
                Console.Write("ALUy=" + CPU.GetRegister(CPURegister.ALUy, Base.hex) + " ");
                Console.Write("ALU=" + CPU.GetRegister(CPURegister.ALUr, Base.hex) + " ");
                Console.Write("USP=" + CPU.GetRegister(CPURegister.USP, Base.hex) + " ");
                Console.Write("SSP=" + CPU.GetRegister(CPURegister.SSP, Base.hex) + " ");
                Console.WriteLine();

                // first line of flags

                // second line of flags
                //Console.WriteLine();
                //Console.WriteLine();
                Console.Write(" IR=" + CPU.GetRegister(CPURegister.IRcmd, Base.hex) + " ");
                Console.Write(" CU=" + CPU.GetRegister(CPURegister.CU, Base.hex) + " ");
                Console.Write("      ");
                Console.Write("S-IE-vvv----VNZC");
                Console.WriteLine();

                Console.Write("INC=" + CPU.GetRegister(CPURegister.INC, Base.hex) + " ");

                Console.Write("         ");
                Console.Write("PSR=" + CPU.GetRegister(CPURegister.PSR, Base.binary));
                Console.WriteLine();

                // opcode representation
                opcode = Memory.Read(CPU.GetRegister(CPURegister.PC));
                opcode = (ushort)(opcode / 256);
                Console.WriteLine(" ---> " + CPU.firmware.getMnemonic(opcode));
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
                return;
            }
            return;
        }

        static void cmdDF(string[] arguments)
        {
            ushort opcode;

            if (arguments.Length == 1)
            {

                Console.Write("PC=" + CPU.GetRegister(CPURegister.PC, Base.hex) + "  ");
                Console.Write("MAR=" + CPU.GetRegister(CPURegister.MAR, Base.hex) + "  ");
                Console.Write("USP=" + CPU.GetRegister(CPURegister.USP, Base.hex) + "  ");
                Console.Write("D0=" + CPU.GetRegister(CPURegister.D0, Base.hex) + "  ");
                Console.Write("D1=" + CPU.GetRegister(CPURegister.D1, Base.hex) + "  ");
                Console.Write("D2=" + CPU.GetRegister(CPURegister.D2, Base.hex) + "  ");
                Console.Write("D3=" + CPU.GetRegister(CPURegister.D3, Base.hex) + "  ");
                Console.WriteLine();

                // next line of registers
                Console.Write("         ");
                Console.Write("MDR=" + CPU.GetRegister(CPURegister.MDR, Base.hex) + "  ");
                Console.Write("SSP=" + CPU.GetRegister(CPURegister.SSP, Base.hex) + "  ");
                Console.Write("D4=" + CPU.GetRegister(CPURegister.D4, Base.hex) + "  ");
                Console.Write("D5=" + CPU.GetRegister(CPURegister.D5, Base.hex) + "  ");
                Console.Write("D6=" + CPU.GetRegister(CPURegister.D6, Base.hex) + "  ");
                Console.Write("D7=" + CPU.GetRegister(CPURegister.D7, Base.hex) + "  ");
                Console.WriteLine();

                // first line of flags

                // second line of flags
                Console.WriteLine();
                Console.Write("IR=" + CPU.GetRegister(CPURegister.IRcmd, Base.hex) + 
                    CPU.GetRegister(CPURegister.IRoperand, Base.hex));
                Console.Write("      ");
                Console.Write(" ALUx=" + CPU.GetRegister(CPURegister.ALUx, Base.hex));
                Console.Write("                    ");
                //Console.Write("S-IE-vvv----VNZC");
                Console.Write("--IE-vvv----VNZC");
                Console.WriteLine();

                Console.Write("                  ");
                Console.Write("ALUy=" + CPU.GetRegister(CPURegister.ALUy, Base.hex) + " ");
                Console.Write("ALU=" + CPU.GetRegister(CPURegister.ALUr, Base.hex));

                Console.Write("       ");
                Console.Write("PSR=" + CPU.GetRegister(CPURegister.PSR, Base.binary));
                Console.WriteLine();

                // opcode representation
                opcode = Memory.Read(CPU.GetRegister(CPURegister.PC));
                opcode = (ushort)(opcode / 256);
                Console.WriteLine();
                Console.WriteLine("Next instruction -> " + CPU.firmware.getMnemonic(opcode));
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
                return;
            }
            return;
        }

        static void old_cmdMD(string[] arguments)
        {
            ushort start = 0;
            ushort stop = 0;
            bool fulldisplay = false;
            bool fileOutput = false;
            string[] separator = { "file=" };
            string[] filenamestring = { };
            ArrayList MDoutput;
            StreamWriter sw;

            string startstring;
            string stopstring;
            string dis;
            string filename;

            if (arguments.Length == 3)
            {
                startstring = arguments[1];
                stopstring = arguments[2];
                dis = "nodis";
                filename = "";

            }
            else if (arguments.Length == 4)
            {
                startstring = arguments[1];
                stopstring = arguments[2];
                dis = arguments[3];
                filename = "";
            }
            else if (arguments.Length == 5)
            {
                startstring = arguments[1];
                stopstring = arguments[2];
                dis = arguments[3];
                filename = arguments[4];
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
                return;
            }

            // do conversions from strings first
            if (supportMethods.isNumber(startstring, ref start) == 0)
            {
                Console.WriteLine("Invalid start address");
                return;
            }

            if (supportMethods.isNumber(stopstring, ref stop) == 0)
            {
                Console.WriteLine("Invalid stop address");
                return;
            }

            if (dis == "dis")
            {
                fulldisplay = true;
            }
            else if (dis == "nodis")
            {
                fulldisplay = false;
            }
            else
            {
                Console.WriteLine("Display status not given or invalid");
                return;
            }

            filenamestring = filename.Split(separator, StringSplitOptions.None);
            if (filename != "")
            {
                if (filenamestring.Length < 2)
                {
                    Console.WriteLine("Filename parameter malformed");
                    return;
                }
                else
                {
                    filename = filenamestring[1];
                    if (filename == "")
                    {
                        Console.WriteLine("Missing filename parameter");
                        return;
                    }
                    else
                    {
                        fileOutput = true;
                    }
                }
            }

            if (fileOutput == true)
            {
                try
                {
                    sw = File.CreateText(filename);
                }
                catch (IOException)
                {
                    Console.WriteLine("File creation failed for some reason, malformed filename?");
                    return;
                }
                MDoutput = MD(start, stop, fulldisplay, ref CPU);
                foreach (string s in MDoutput)
                {
                    sw.WriteLine(s);
                }
                sw.Close();

            }
            else
            {
                MDoutput = MD(start, stop, fulldisplay, ref CPU);
                foreach (string s in MDoutput)
                {
                    Console.WriteLine(s);
                }
            }
            return;
        }

        static void cmdMD(string[] arguments)
        {
            ushort start = 0;
            ushort stop = 0;
            bool fulldisplay = false;
            bool fileOutput = false;
            string[] separator = { "file=" };
            string[] filenamestring = { };
            ArrayList MDoutput;
            StreamWriter sw;

            int length = 10; // default length

            string startstring;
            string lengthstring;
            string dis;
            string filename;

            if (arguments.Length == 2)
            {
                startstring = arguments[1];
                lengthstring = length.ToString();
                dis = "nodis";
                filename = "";

            }
            else if (arguments.Length == 3)
            {
                startstring = arguments[1];
                lengthstring = arguments[2];
                dis = "nodis";
                filename = "";

            }
            else if (arguments.Length == 4)
            {
                startstring = arguments[1];
                lengthstring = arguments[2];
                dis = arguments[3];
                filename = "";
            }
            else if (arguments.Length == 5)
            {
                startstring = arguments[1];
                lengthstring = arguments[2];
                dis = arguments[3];
                filename = arguments[4];
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
                return;
            }

            // do conversions from strings first
            if (supportMethods.isNumber(startstring, ref start) == 0)
            {
                Console.WriteLine("Invalid start address");
                return;
            }

            if (supportMethods.isNumber(lengthstring, ref stop) == 0)
            {
                Console.WriteLine("Invalid length");
                return;
            }

            if (dis == "dis")
            {
                fulldisplay = true;
            }
            else if (dis == "nodis")
            {
                fulldisplay = false;
            }
            else
            {
                Console.WriteLine("Display status not given or invalid");
                return;
            }

            filenamestring = filename.Split(separator, StringSplitOptions.None);
            if (filename != "")
            {
                if (filenamestring.Length < 2)
                {
                    Console.WriteLine("Filename parameter malformed");
                    return;
                }
                else
                {
                    filename = filenamestring[1];
                    if (filename == "")
                    {
                        Console.WriteLine("Missing filename parameter");
                        return;
                    }
                    else
                    {
                        fileOutput = true;
                    }
                }
            }

            if (fileOutput == true)
            {
                try
                {
                    sw = File.CreateText(filename);
                }
                catch (IOException)
                {
                    Console.WriteLine("File creation failed for some reason, malformed filename?");
                    return;
                }
                MDoutput = MDlength(start, stop, fulldisplay, ref CPU);
                foreach (string s in MDoutput)
                {
                    sw.WriteLine(s);
                }
                sw.Close();

            }
            else
            {
                MDoutput = MDlength(start, stop, fulldisplay, ref CPU);
                foreach (string s in MDoutput)
                {
                    Console.WriteLine(s);
                }
            }
            return;
        }

        static void cmdJasm(string[] arguments)
        {
            Console.WriteLine("Assembler testbed not yet implemented");
            return;
        }

        static void cmdJCC(string[] arguments)
        {

            Console.WriteLine("Compiler testbed not yet implemented");
            return;

            //ArrayList tokens;
            //ArrayList source;

            //if (arguments.Length == 2)
            //{
            //    if (validatedFilename(ref arguments[1]))
            //    {
            //        source = Compiler.LoadSrcFile(arguments[1]);
            //        tokens = Compiler.LexicalAnalysis(source);


            //        //tokens = Compiler.ModifyTokens(tokens);
            //        //Compiler.ListTokens(tokens);
            //        //tagtokens = Compiler.Convert(tokens);
            //        //Compiler.ListTagTokens(tagtokens);

                    
            //    }
            //}
            //else
            //{
            //    MessageLog.AddMessage(messagesType.Error, 0, "Too many params to help");
            //}
            //return;
        }

        static void cmdRegister(CPURegister register, string[] arguments)
        {
            ushort result = 0;

            if (arguments.Length == 1)
            {
                Console.WriteLine(register + "=" 
                    + CPU.GetRegister(register, Base.hex)
                    + " ("
                    + CPU.GetRegister(register, Base.binary)
                    + ")"
                    );
                //Console.WriteLine();
            }
            else if (arguments.Length == 2)
            {
                if (supportMethods.isNumber(arguments[1], ref result) == 0)
                {
                    Console.WriteLine("Number format invalid");
                }
                else
                {
                    CPU.SetRegister(register, result);
                    //checkMessageStore(CPU);
                    showMessages();

                }
            }
            else
            {
                Console.WriteLine("Incorrect number of parameters on command line");
            }
            return;
        }

        static void cmdBF(string[] arguments)
        {
            ushort start = 0;
            ushort stop = 0;
            ushort value = 0;
            
            if (arguments.Length != 4)
            {
                Console.WriteLine("Incorrect number of parameters on command line");
                return;
            }

            // do conversions from strings first
            if (supportMethods.isNumber(arguments[1], ref start) == 0)
            {
                Console.WriteLine("Invalid start address");
                return;
            }

            if (supportMethods.isNumber(arguments[2], ref stop) == 0)
            {
                Console.WriteLine("Invalid stop address");
                return;
            }

            if (supportMethods.isNumber(arguments[3], ref value) == 0)
            {
                Console.WriteLine("Invalid data word for value");
                return;
            }

            BF(start, stop, value);
            //checkMessageStore(MemoryManager);
            return;
        }

        static void cmdBM(string[] arguments)
        {

            ushort start = 0;
            ushort stop = 0;
            ushort target = 0;

            if (arguments.Length != 4)
            {
                Console.WriteLine("Incorrect number of parameters on command line");
                return;
            }

            // do conversions from strings first
            if (supportMethods.isNumber(arguments[1], ref start) == 0)
            {
                Console.WriteLine("Invalid start address");
                return;
            }

            if (supportMethods.isNumber(arguments[2], ref stop) == 0)
            {
                Console.WriteLine("Invalid stop address");
                return;
            }

            if (supportMethods.isNumber(arguments[3], ref target) == 0)
            {
                Console.WriteLine("Invalid target address");
                return;
            }

            BM(start, stop, target);
            //checkMessageStore(MemoryManager);
            return;
        }

        //--------------------------------------------------------------------
        static void cmdCD(string[] arguments)
        {
            string cwd = "";

            if (arguments.Length == 1)
            {
                cwd = Directory.GetCurrentDirectory();
                Console.WriteLine(cwd);
            }
            else if (arguments.Length == 2)
            {
                try
                {
                    Directory.SetCurrentDirectory(arguments[1]);
                }
                catch (IOException)
                {
                    Console.WriteLine("Could not set directory to this");
                }
                finally
                {
                }

            }
            else
            {
                Console.WriteLine("Wrong number of params");
            }
            return;
        }


        //--------------------------------------------------------------------
        static void cmdLS()
        {

            // get current directory
            string cwd = Directory.GetCurrentDirectory();

            string[] files = Directory.GetFiles(cwd);
            string[] directories = Directory.GetDirectories(cwd);

            Console.WriteLine();
            Console.WriteLine(" Directory of " + cwd);
            Console.WriteLine();

            // list directories
            foreach (string directory in directories)
            {
                string filename = System.IO.Path.GetFileName(directory);
                Console.WriteLine("<DIR>  " + filename);
            }

            // list files
            foreach (string file in files)
            {
                string filename = System.IO.Path.GetFileName(file);
                Console.WriteLine("       " + filename);
            }
        }

        //--------------------------------------------------------------------
        static void cmdEnvLS()
        {

            // get current directory
            string cwd = Environment.GetEnvironmentVariable(globalConstants.envName);

            if (cwd != null)
            {
                if (Directory.Exists(cwd))
                {
                    string[] files = Directory.GetFiles(cwd);
                    string[] directories = Directory.GetDirectories(cwd);

                    Console.WriteLine();
                    Console.WriteLine(" Environment directory of " + cwd);
                    Console.WriteLine();

                    // list directories
                    foreach (string directory in directories)
                    {
                        string filename = System.IO.Path.GetFileName(directory);
                        Console.WriteLine("<DIR>  " + filename);
                    }

                    // list files
                    foreach (string file in files)
                    {
                        string filename = System.IO.Path.GetFileName(file);
                        Console.WriteLine("       " + filename);
                    }
                }
                else
                {
                    Console.WriteLine(globalConstants.envName + " environment variable is not set to a valid location");
                }

            }
            else
            {
                Console.WriteLine(globalConstants.envName + " environment variable is not set");

            }
        }

        static string convertSynonyms(string cmd)
        {
            string returnValue = cmd;
            switch (cmd)
            {
                case "?":
                case "he":
                case "hel":
                case "help":
                    returnValue = "help";
                    break;
                case "ver":
                case "vers":
                case "versi":
                case "versio":
                case "version":
                    returnValue = "version";
                    break;
                case "res":
                case "rese":
                case "reset":
                    returnValue = "reset";
                    break;
                case "f":
                case "fr":
                    returnValue = "frm";
                    break;
                case "l":
                case "lo":
                case "loa":
                case "load":
                    returnValue = "load";
                    break;
                case "go":
                case "run":
                    returnValue = "run";
                    break;
                case "t":
                case "tr":
                case "tra":
                case "trac":
                case "trace":
                    returnValue = "trace";
                    break;
                case "ab":
                case "abo":
                case "abou":
                case "about":
                    returnValue = "about";
                    break;
                case "fe":
                case "fet":
                case "fetc":
                case "fetch":
                    returnValue = "fetch";
                    break;
                case "dir":
                    returnValue = "ls";
                    break;
                case "envdir":
                    returnValue = "envls";
                    break;
                case "exec":
                case "execu":
                case "execut":
                case "execute":
                    returnValue = "execute";
                    break;
                case "ju":
                case "jum":
                case "jump":
                case "jumpe":
                case "jumper":
                case "jumpers":
                    returnValue = "jumpers";
                    break;
                case "mem":
                case "memu":
                case "memup":
                case "memupd":
                case "memupda":
                case "memupdat":
                case "memupdate":
                    returnValue = "memupdate";
                    break;
                case "per":
                case "peri":
                case "perip":
                case "periph":
                    returnValue = "periph";
                    break;
                case "io":
                case "iob":
                case "iobo":
                case "ioboa":
                case "ioboar":
                case "ioboard":
                    returnValue = "ioboard";
                    break;
                case "ex":
                case "exi":
                case "exit":
                case "q":
                case "qu":
                case "qui":
                case "quit":
                    returnValue = "quit";
                    break;
                case "/?":
                case "/h":
                case "/he":
                case "/hel":
                case "/help":
                    returnValue = "/help";
                    break;
                case "/g":
                case "/go":
                    returnValue = "/go";
                    break;
                case "/q":
                case "/qu":
                case "/qui":
                case "/quit":
                    returnValue = "/quit";
                    break;
                case "/l":
                case "/lo":
                case "/loa":
                case "/load":
                    returnValue = "/load";
                    break;
                case "/f":
                case "/fr":
                    returnValue = "/frm";
                    break;
                case "/n":
                case "/no":
                case "/nod":
                case "/node":
                case "/nodef":
                case "/nodefa":
                case "/nodefau":
                case "/nodefaul":
                case "/nodefault":
                    returnValue = "/nodefault";
                    break;
                case "clear":
                case "clr":
                    returnValue = "cls";
                    break;
                default:
                    break;
            }
            return (returnValue);
        }

        //static bool removed_validatedFilename(ref string filename)
        //{
        //    bool fileExists = false;
        //    string error = "";

        //    string env = "";
        //    string updatedfilename = "";

        //    if (File.Exists(filename))
        //    {
        //        fileExists = true;
        //    }
        //    else
        //    {
        //        error = "File [" + filename + "] does not exist in local directory";
        //        Console.WriteLine(error);
        //        env = Environment.GetEnvironmentVariable(globalConstants.envName);
        //        if (env != "")
        //        {
        //            if (Directory.Exists(env))
        //            {
        //                updatedfilename = env + filename;
        //                if (File.Exists(updatedfilename))
        //                {
        //                    error = "Loading file [" + filename + "] from " + globalConstants.envName + " directory";
        //                    Console.WriteLine(error);
        //                    filename = updatedfilename;
        //                    fileExists = true;
        //                }
        //                else
        //                {
        //                    error = "File [" + filename + "] does not exist in " + globalConstants.envName + " directory";
        //                    Console.WriteLine(error);
        //                }
        //            }
        //            else
        //            {
        //                error = "Cannot load [" + filename + "] as " + globalConstants.envName + " environment directory does not exist";
        //                Console.WriteLine(error);
        //            }
        //        }
        //    }
        //    return (fileExists);
        //}

        static void HardReset()
        {
            CPU.HardReset();
            showMessages();

            Memory.HardReset();
            SetNMI(ref CPU);
            return;
        }

        static void Help()
        {
            Console.WriteLine();
            Console.WriteLine("  " + programName + " [/h*elp]");
            Console.WriteLine("   or");
            Console.WriteLine("  " + programName + " [/l*oad={machinecodefile}] [/f*rm={firmwarefile}]");
            Console.WriteLine("        [/g*o] [/q*uit] [/pc={addr}] ] [/n*odefault]");
            Console.WriteLine();
            Console.WriteLine("  After entering " + programName + ", type 'HELP' for more information.");
            Console.WriteLine();
            Environment.Exit(0);   
        }

        static void LoadProgram(string filename, ref CPU cpu)
        {
            string contents = "";
            string[] lines = { };
            string[] line = { };
            string[] separator = { Environment.NewLine };
            string[] spaceseparator = { " " };

            string error;

            ushort address = 0;
            ushort value = 0;

            int lineNumber = 0;
            string message = "";

            if (File.Exists(filename))
            {
                // read the file
                contents = File.ReadAllText(filename);

                lines = contents.Split(separator, StringSplitOptions.None);

                for (int count = 0; count < lines.Length; count++)
                {
                    line = lines[count].Split(spaceseparator, StringSplitOptions.RemoveEmptyEntries);

                    if (line.Length > 0)
                    {
                        if (line[0].Equals("org", StringComparison.CurrentCultureIgnoreCase))
                        {
                            if (line.Length > 1)
                            {
                                // get param 2 for address
                                if (supportMethods.isNumber(line[1], ref value) == 0)
                                {
                                    lineNumber = count + 1;
                                    message = "Invalid data value at line " + lineNumber;
                                    //MessageStore.AddMessage(messagesType.Error, 1, message);
                                    //return;
                                }
                                else
                                {
                                    // set address to value of ORG
                                    address = value;
                                }
                            }
                            else
                            {
                                lineNumber = count + 1;
                                message = "Invalid data value at line " + lineNumber;
                                //MessageStore.AddMessage(messagesType.Error, 1, message);
                                //return;
                            }
                        }
                        else if (line[0].Equals("lpc", StringComparison.CurrentCultureIgnoreCase))
                        {
                            if (line.Length > 1)
                            {
                                // get param 2 for address
                                if (supportMethods.isNumber(line[1], ref value) == 0)
                                {
                                    lineNumber = count + 1;
                                    message = "Invalid data value at line " + lineNumber;
                                    //MessageStore.AddMessage(messagesType.Error, 201, message);
                                    //return;
                                }
                                else
                                {
                                    // set address to value of PC
                                    cpu.SetRegister(CPURegister.PC, value);
                                    message = "PC=" + cpu.GetRegister(CPURegister.PC, Base.hex);
                                    //MessageStore.AddMessage(messagesType.Information, 201, message);
                                    //showMessages();

                                }
                            }
                            else
                            {
                                lineNumber = count + 1;
                                message = "Invalid data value at line " + lineNumber;
                                //MessageStore.AddMessage(messagesType.Error, 1, message);
                                //return;
                            }
                        }
                        else if (line[0].Equals("32bit", StringComparison.CurrentCultureIgnoreCase))
                        {
                            message = "Can't currently load 32-bit memory dumps";
                            //MessageStore.AddMessage(messagesType.Warning, 2, message);
                            return;
                        }
                        else if (line[0].Substring(0, 1) == "#")
                        {
                            // comment
                        }
                        else if (line[0].Substring(0, 1) == "*")
                        {
                            // comment
                        }
                        else
                        {
                            // get param 2 for address
                            if (supportMethods.isNumber(line[0], ref value) == 0)
                            {
                                lineNumber = count + 1;
                                message = "Invalid data value at line " + lineNumber;
                                //MessageStore.AddMessage(messagesType.Error, 1, message);
                                //return;
                            }
                            else
                            {
                                // put value into memory at address then increment address
                                Memory.Write(address, value);
                                address++;
                            }
                        }
                    }
                }
            }
            else
            {
                showMessages();

                error = "File [" + filename + "] does not exist in local or environment directory";
                Console.WriteLine(error);

            }

            return;
        }

        static void BM(ushort start, ushort stop, ushort target)
        {
            // check stop is greater than start
            if (start > stop)
            {
                //MessageStore.AddMessage(messagesType.Error, 5, "Start address is greater than stop address");
                return;
            }

            if ((target >= start) && (target <= stop))
            {
                //MessageStore.AddMessage(messagesType.Error, 6, "Target inside move area");
                return;
            }

            // do the block move
            for (int i = start; i <= stop; i++)
            {
                Memory.Write(target, Memory.Read((ushort)(i)));
                target++;
            }

            return;
        }

        // this should be removed from here and placed in the main program
        static void BF(ushort start, ushort stop, ushort value)
        {
            // check stop is greater than start
            if (start > stop)
            {
                //MessageStore.AddMessage(messagesType.Error, 7, "Start address is greater than stop address");
                return;
            }

            // do the block file
            for (int i = start; i <= stop; i++)
            {
                Memory.Write((ushort)i, value);
            }
            return;
        }

        static ArrayList MDlength(ushort start, ushort length, bool dis, ref CPU CPU)
        {
            string s = "";
            ushort displayaddress = 0;
            int count = 0;

            int intstop = 0;
            ushort stop = 0;

            intstop = start + length - 1;

            if (intstop > 0xffff) {
                stop = (ushort)0xffff;
            }
            else {
                stop = (ushort)intstop;
            }

            // storage for the MD
            ArrayList array = new ArrayList();

            //int index = 0;

            // check stop is greater than start
            if (start > stop)
            {
                //MessageStore.AddMessage(messagesType.Error, 3, "Start address is greater than stop address");

            }
            else
            {

                if (dis)
                {
                    array.Add("org " + supportMethods.FormatUshort(start, Base.hex));

                    for (count = start; count <= stop; count++)
                    {
                        s = supportMethods.FormatUshort(Memory.Read((ushort)count), Base.hex);
                        s = s + "  #  ";
                        s = s + supportMethods.FormatUshort((ushort)count, Base.hex);
                        if (((count) % 2) == 0) // if it's an even address, show the mnemonic
                        {
                            s = s + "  " + CPU.firmware.getMnemonic((ushort)(Memory.Read((ushort)count) / 256));
                        }
                        array.Add(s);
                        //index++;
                    }

                }
                else
                {
                    displayaddress = start;
                    for (count = start; count <= stop; count++)
                    {
                        if (((count) % 8) == 0)
                        {
                            s = supportMethods.FormatUshort((ushort)displayaddress, Base.hex) + " : ";
                        }

                        s = s + supportMethods.FormatUshort(Memory.Read((ushort)count), Base.hex) + " ";

                        if (((count + 1) % 8) == 0)
                        {
                            array.Add(s);
                            //index++;
                        }
                        displayaddress++;
                    }
                    if ((count % 8) != 0)
                    {
                        array.Add(s);
                        //index++;
                    }

                }

            }
            return (array);
        }

        static ArrayList MD(ushort start, ushort stop, bool dis, ref CPU CPU)
        {
            string s = "";
            ushort displayaddress = 0;
            int count = 0;

            // storage for the MD
            ArrayList array = new ArrayList();

            //int index = 0;

            // check stop is greater than start
            if (start > stop)
            {
                //MessageStore.AddMessage(messagesType.Error, 3, "Start address is greater than stop address");

            }
            else
            {
                array.Add("org " + supportMethods.FormatUshort(start, Base.hex));

                if (dis)
                {
                    for (count = start; count <= stop; count++)
                    {
                        s = supportMethods.FormatUshort(Memory.Read((ushort)count), Base.hex);
                        s = s + "  #  ";
                        s = s + supportMethods.FormatUshort((ushort)count, Base.hex);
                        if (((count) % 2) == 0) // if it's an even address, show the mnemonic
                        {
                            s = s + "  " + CPU.firmware.getMnemonic((ushort)(Memory.Read((ushort)count) / 256));
                        }
                        array.Add(s);
                        //index++;
                    }

                }
                else
                {
                    displayaddress = start;
                    for (count = start; count <= stop; count++)
                    {
                        if (((count) % 8) == 0)
                        {
                            s = supportMethods.FormatUshort((ushort)displayaddress, Base.hex) + " : ";
                        }

                        s = s + supportMethods.FormatUshort(Memory.Read((ushort)count), Base.hex) + " ";

                        if (((count + 1) % 8) == 0)
                        {
                            array.Add(s);
                            //index++;
                        }
                        displayaddress++;
                    }
                    if ((count % 8) != 0)
                    {
                        array.Add(s);
                        //index++;
                    }

                }

            }
            return (array);
        }

        static void SetNMI(ref CPU CPU)
        {
            // install the default NMI handler
            // point to first address in ROM, ROM laoded by default at 0x0000
            Memory.Write((ushort)(CPU.InterruptBase + 7), 0x0000); 
            return;
        }

        static void showMessages() {
            if (messageList.Count > 0)
            {
                ArrayList a = messageList.Retrieve();
                for (int i = 0; i < a.Count; i++ )
                {
                    Console.WriteLine((string)a[i]);
                }
            }
        }


    } // end class
}
    