//Copyright (c) 2009, Mark S. Burrell
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are 
//permitted provided that the following conditions are met:

//    * Redistributions of source code must retain the above copyright notice, this list 
//of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this 
//list of conditions and the following disclaimer in the documentation and/or other 
//materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
//EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
//OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
//SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
//OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;

using Sedici.Hardware;

namespace Sedici.Hardware
{
    public class SediciAssembler
    {


        // storage for the original assembler source file
        private ArrayList srcFile;          // storage for original input file - an array of strings
        private ArrayList tokenSet;         // storage for resulting tokens - an array of ArrayLists

        private Hashtable SymbolTable;

        private ArrayList intermediateMachineCode;

        private Firmware firmware;

        private string finalMachineCode;

        public string Name;
        public string InstructionSet;
        public string assemblyfile;

        public int Errors;

        private bool outputTypeMachineCode;


        public SediciAssembler() {
            srcFile = new ArrayList();
            tokenSet = new ArrayList();

            SymbolTable = new Hashtable();

            intermediateMachineCode = new ArrayList();

            firmware = new Firmware();

            Name = "Sedici Assembler";

            outputTypeMachineCode = false; // default is pretty print

            Errors = 0;
        }

        public void setOutputToMachineCode (bool b) {
            outputTypeMachineCode = b;
        }

        public void loadFirmware(string filename)
        {
            firmware.LoadFirmware(filename);
            InstructionSet = filename;

        }

        public void loadAssemblyCode(string filename) {
            srcFile.Clear();

            // load the file


            srcFile = LoadAssemblerSourceFile(filename);

            assemblyfile = filename;


            // process the USE directives
            int Count = 0;

            while (foundValidUSE(srcFile))
            {

                srcFile = includeUSE(srcFile);
                Count++;

                if (Count == 100)
                {
                    //MessageBox.Show("Error : USEs in a loop or going to infinity - stopped");
                    break;
                }
            }


        }

        public string getTokenSet() {
            string Text = "";

            ArrayList lineTokens;

            // update the text field
            Text = "";

            for (int i = 0; i < tokenSet.Count; i++)
            {
                lineTokens = (ArrayList)tokenSet[i];


                for (int j = 0; j < lineTokens.Count; j++)
                {
                    Text = Text + "[" + lineTokens[j] + "]";
                }
                Text = Text + Environment.NewLine;

            }

            return (Text);
        }

        public string getIntermediateMachineCode() {

            ArrayList lineTokens;
            string Text;

            // display the intermediateMachineCode array
            Text = "";
            for (int i = 0; i < intermediateMachineCode.Count; i++)
            {
                lineTokens = (ArrayList)intermediateMachineCode[i];

                if (lineTokens.Count > 0)
                {
                    if (!(lineTokens[0].Equals("")))
                    {
                        Text = Text + "[" + lineTokens[0] + "]";
                        Text = Text + Environment.NewLine;
                    }
                }
            }

            return(Text);

        }

        public string getSymbolTable()
        {
            string Text;

            Text = "Symbols" + Environment.NewLine + "-------" + Environment.NewLine;

            ArrayList symbolList = new ArrayList();

            foreach (DictionaryEntry entry in SymbolTable)
            {
                symbolList.Add(entry.Key);
            }

            symbolList.Sort();

            int lineLength = 10;

            foreach (string listentry in symbolList)
            {
                ushort eric = (ushort)SymbolTable[listentry];


                string fred = supportMethods.FormatUshort(eric, Base.hex).ToUpper();

                Text = Text + "#  " + listentry + " ";

                for (int z = 0; z < (lineLength - listentry.Length); z++)
                {
                    Text = Text + " ";
                }

                Text = Text + " " + fred;
                Text = Text + Environment.NewLine;

            }

            return (Text);
        }

        public void initialTokenisation()
        {

            string Text;

            ArrayList lineTokens;

            // update the public tokenSet with the contents of the array of strings from the file
            tokenSet.Clear();
            tokenSet = writeInitialTokenSet(srcFile);

            tokenSet = writeFinalTokenSet(tokenSet);

            Boolean doDisplay = false;

            if (doDisplay)
            {


                // update the text field
                Text = "";

                for (int i = 0; i < tokenSet.Count; i++)
                {
                    lineTokens = (ArrayList)tokenSet[i];

                    if (lineTokens.Count == 4)
                    {
                        // all ok
                        if ((lineTokens[0].ToString() == "") &&
                            (lineTokens[1].ToString() == "") &&
                            (lineTokens[2].ToString() == ""))
                        {
                            // blank line
                            Text = Text +
                                "        * " +
                                lineTokens[3] + Environment.NewLine;
                        }
                        else if (lineTokens[0].ToString() == "")
                        {
                            Text = Text +
                                "        * " +
                                "\t" + lineTokens[1] +
                                "\t" + lineTokens[2] + Environment.NewLine;
                        }
                        else
                        {
                            Text = Text +
                                "        * " +
                                lineTokens[0] +
                                "\t" + lineTokens[1] +
                                "\t" + lineTokens[2] + Environment.NewLine;
                        }

                    }
                    else if (lineTokens.Count == 5)
                    {
                        // error stored
                        Errors++;
                        Text = Text + "* Error * "
                            + lineTokens[4] + Environment.NewLine;
                    }
                    else
                    {
                        //error in processing here
                        Errors++;
                        Text = Text + "------->ERROR" + Environment.NewLine;
                    }


                }


            }


        }

        public void initialCoding()
        {

            intermediateMachineCode.Clear();
            SymbolTable.Clear();

            ArrayList intermediate;

            String label = "";
            String instruction = "";
            String data = "";
            String line = "";

            ArrayList lineTokens;

            // update the text field
            //Text = "";

            for (int i = 0; i < tokenSet.Count; i++)
            {
                lineTokens = (ArrayList)tokenSet[i];

                // ignore any errors already found
                if (lineTokens.Count == 4)
                {
                    label = lineTokens[0].ToString();
                    instruction = lineTokens[1].ToString();
                    data = lineTokens[2].ToString();
                    line = lineTokens[3].ToString();

                    if ((instruction == "") && (label == ""))
                    {
                        intermediate = processBlank((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                    else if ((instruction == "") && (label != ""))
                    {
                        // deal with any labels
                        if ((!label.Equals("")))
                        {
                            //Text = Text + "setsymbol|" + label + ":";
                            intermediate = processSymbol((ArrayList)tokenSet[i]);
                            intermediateMachineCode.Add(intermediate);
                        }
                    }
                    else if (string.Equals(instruction, "equ", StringComparison.CurrentCultureIgnoreCase))
                    {
                        intermediate = processEQU((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                    else if (string.Equals(instruction, "dc.b", StringComparison.CurrentCultureIgnoreCase))
                    {
                        intermediate = processDCB((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                    else if (string.Equals(instruction, "ds.w", StringComparison.CurrentCultureIgnoreCase))
                    {
                        intermediate = processDSW((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                    else if (string.Equals(instruction, "dc.w", StringComparison.CurrentCultureIgnoreCase))
                    {
                        intermediate = processDCW((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                    else if (string.Equals(instruction, "org", StringComparison.CurrentCultureIgnoreCase))
                    {
                        intermediate = processORG((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                    else if (string.Equals(instruction, "lpc", StringComparison.CurrentCultureIgnoreCase))
                    {
                        intermediate = processLPC((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                    else
                    {
                        // it is an instruction
                        intermediate = processInstruction((ArrayList)tokenSet[i]);
                        intermediateMachineCode.Add(intermediate);
                    }
                }
                else if (lineTokens.Count == 5)
                {
                    // error in line
                    intermediate = processError((ArrayList)tokenSet[i]);
                    intermediateMachineCode.Add(intermediate);

                }

            }
        }

        public string prettyPrint()
        {
            string Text;
            ArrayList intermediateTokens;

            string[] separator = { ":" };
            string[] barseparator = { "|" };

            string[] lineEntries;
            string[] keypairs;

            ushort number = 0;

            int lineNumber = 0;

            string errorString = "";

            int errorCount = 0;

            ushort address = 0;

            Text = "";
            Text = Text + "#-------------------------------------------------------------------------------" + Environment.NewLine;
            Text = Text + "# " + globalConstants.COPYRIGHT + Environment.NewLine;
            Text = Text + "# " + globalConstants.VERSION + Environment.NewLine;
            Text = Text + "#" + Environment.NewLine;
            Text = Text + "# Assembler input : " + assemblyfile + Environment.NewLine;
            Text = Text + "# Instruction set : " + InstructionSet + Environment.NewLine;
            Text = Text + "#" + Environment.NewLine;
            Text = Text + "#-------------------------------------------------------------------------------" + Environment.NewLine;
            Text = Text + "#" + Environment.NewLine;
            Text = Text + "# Assembler Listing" + Environment.NewLine;
            Text = Text + "#" + Environment.NewLine;
            Text = Text + "#-------------------------------------------------------------------------------" + Environment.NewLine;

            for (int i = 0; i < intermediateMachineCode.Count; i++)
            {

                intermediateTokens = (ArrayList)intermediateMachineCode[i];

                //for (int j = 0; j < (intermediateTokens.Count); j++)
                //{
                if (intermediateTokens[0].ToString().Equals(""))
                {
                    // nothing on the line
                    Text = Text + "          # " + 
                        supportMethods.FormatUshort(address, Base.hex).ToUpper() 
                        + " # " + intermediateTokens[1].ToString() 
                        + Environment.NewLine;
                }
                else
                {
                    lineEntries = intermediateTokens[0].ToString().Split(separator, StringSplitOptions.None);

                    lineNumber = 0;

                    for (int k = 0; k < lineEntries.Length; k++)
                    {
                        keypairs = lineEntries[k].Split(barseparator, StringSplitOptions.None);
                        switch (keypairs[0])
                        {
                            case "lpc":
                                supportMethods.isNumber(keypairs[1], ref number);
                                Text = Text + "LPC $" + supportMethods.FormatUshort(number, Base.hex).ToUpper();
                                Text = Text + " # "
                                    + supportMethods.FormatUshort(address, Base.hex).ToUpper() + " # ";

                                break;
                            case "org":
                                supportMethods.isNumber(keypairs[1], ref number);
                                Text = Text + "ORG $" + supportMethods.FormatUshort(number, Base.hex).ToUpper();
                                address = number;
                                Text = Text + " # "
                                    + supportMethods.FormatUshort(address, Base.hex).ToUpper() + " # ";

                                break;
                            case "actual":
                                //case "opcode":
                                supportMethods.isNumber(keypairs[1], ref number);
                                Text = Text + supportMethods.FormatUshort(number, Base.hex).ToUpper() + "     ";
                                Text = Text + " # "
                                    + supportMethods.FormatUshort(address, Base.hex).ToUpper() + " # ";
                                address++;

                                break;
                            case "error":
                                errorCount++;
                                Text = Text + "         ";
                                Text = Text + " # "
                                    + supportMethods.FormatUshort(address, Base.hex).ToUpper() + " # ";


                                if (errorString.Equals(""))
                                {
                                    errorString = errorString + "# Error : " + keypairs[1];
                                }
                                else
                                {
                                    errorString = errorString + Environment.NewLine + "# Error : " + keypairs[1];
                                }
                                break;
                            default:
                                // should never happen
                                Text = Text + "         ";
                                Text = Text + " # "
                                    + supportMethods.FormatUshort(address, Base.hex).ToUpper() + " # ";
                                Text = Text + "[Unprocessed tag : " + keypairs[0] + "] ";
                                break;
                        }

                        if (lineNumber == 0)
                        {
                            Text = Text + intermediateTokens[1].ToString() + Environment.NewLine;
                        }
                        else
                        {
                            Text = Text + Environment.NewLine;
                        }

                        if (!(errorString.Equals("")))
                        {

                            Text = Text + errorString + Environment.NewLine + "#" + Environment.NewLine;
                            errorString = "";
                        }

                        lineNumber++;
                    }


                }

                //}


            }

            // now show symbol table
            //            Text = Text + "#\r\n";
            //            Text = Text + "# Symbols\r\n#\r\n";

            Text = Text + "#-------------------------------------------------------------------------------" + Environment.NewLine;
            Text = Text + "#" + Environment.NewLine;
            Text = Text + "# Symbol table" + Environment.NewLine;
            Text = Text + "#" + Environment.NewLine;
            Text = Text + "#-------------------------------------------------------------------------------" + Environment.NewLine;

            ArrayList symbolList = new ArrayList();
            foreach (DictionaryEntry entry in SymbolTable)
            {
                symbolList.Add(entry.Key);
            }

            symbolList.Sort();

            int lineLength = 40;

            foreach (string listentry in symbolList)
            {
                ushort eric = (ushort)SymbolTable[listentry];


                string fred = supportMethods.FormatUshort(eric, Base.hex).ToUpper();

                Text = Text + "#  " + listentry + " ";

                for (int z = 0; z < (lineLength - listentry.Length); z++)
                {
                    Text = Text + ".";
                }

                Text = Text + " " + fred;
                Text = Text + Environment.NewLine;

            }


            if (errorCount > 0)
            {
                Errors = errorCount;

                Text = Text + "#" + Environment.NewLine + "#" + Environment.NewLine + "# Assembly completed with ";
                if (errorCount == 1)
                {
                    Text = Text + "1 error" + Environment.NewLine + "#" + Environment.NewLine;
                }
                else
                {
                    Text = Text + errorCount + " errors" + Environment.NewLine + "#" + Environment.NewLine;
                }
            }
            else
            {
                Text = Text + "#" + Environment.NewLine 
                    + "#" + Environment.NewLine + "# Assembly completed successfully"
                    + Environment.NewLine + "#" + Environment.NewLine;
            }

            return (Text);

        }

        public string prettyPrintMachineCode()
        {
            string Text;
            ArrayList intermediateTokens;

            string[] separator = { ":" };
            string[] barseparator = { "|" };

            string[] lineEntries;
            string[] keypairs;

            ushort number = 0;

            Text = "";

            for (int i = 0; i < intermediateMachineCode.Count; i++)
            {

                intermediateTokens = (ArrayList)intermediateMachineCode[i];

                if (intermediateTokens[0].ToString().Equals(""))
                {
                }
                else
                {
                    lineEntries = intermediateTokens[0].ToString().Split(separator, StringSplitOptions.None);

                    for (int k = 0; k < lineEntries.Length; k++)
                    {
                        keypairs = lineEntries[k].Split(barseparator, StringSplitOptions.None);
                        switch (keypairs[0])
                        {
                            case "lpc":
                                supportMethods.isNumber(keypairs[1], ref number);
                                Text = Text + "LPC $" + supportMethods.FormatUshort(number, Base.hex).ToUpper();
                                Text = Text + Environment.NewLine;
                                break;
                            case "org":
                                supportMethods.isNumber(keypairs[1], ref number);
                                Text = Text + "ORG $" + supportMethods.FormatUshort(number, Base.hex).ToUpper();
                                Text = Text + Environment.NewLine;
                                break;
                            case "actual":
                                supportMethods.isNumber(keypairs[1], ref number);
                                Text = Text + supportMethods.FormatUshort(number, Base.hex).ToUpper();
                                Text = Text + Environment.NewLine;
                                break;
                            case "error":
                                Errors++;
                                break;
                            default:
                                break;
                        }
                    }
                }
            }


            return (Text);

        }


        public void processDSWReferences()
        {
            // replace dsw references
            // can be in form like dsw|4 
            // or dsw|symbol (where the symbol is already in the symbol table)

            ArrayList lineTokens;
            ArrayList newarray;
            string[] separator = { ":" };
            string[] token = { };

            string updated = "";
            string line = "";
            bool dsw_found = false;

            string dswstring = "";
            ushort number = 0;

            newarray = new ArrayList();

            // go through each line
            for (int i = 0; i < intermediateMachineCode.Count; i++)
            {
                // on each line we have a set of instructions
                lineTokens = (ArrayList)intermediateMachineCode[i];

                // split the instructions based on the : character
                token = lineTokens[0].ToString().Split(separator, StringSplitOptions.RemoveEmptyEntries);
                line = lineTokens[1].ToString();

                updated = "";
                dsw_found = false;

                // go through the instructions looking for dsw| , any other entry simply left alone
                for (int j = 0; j < token.Length; j++)
                {
                    if (token[j].StartsWith("dsw|"))
                    {
                        // we have found a dsw reference, now what sort is it?
                        dsw_found = true;

                        string local = token[j];
                        dswstring = local.Substring(4, (local.Length - 4));

                        if (supportMethods.isNumber(dswstring, ref number) == 1)
                        {
                            for (int z = 0; z < number; z++)
                            {
                                updated = addString(updated, "actual|$0000");
                            }
                        }
                        else
                        {
                            // not a number so see if its a symbol reference
                            if (dswstring.StartsWith("#"))
                            {
                                // then get the symbol and look it up
                                string symbol = dswstring.Substring(1, (dswstring.Length - 1));

                                try
                                {
                                    number = (ushort)SymbolTable[symbol];

                                    for (int z = 0; z < number; z++)
                                    {
                                        updated = addString(updated, "actual|$0000");
                                    }
                                }
                                catch (Exception)
                                {
                                    updated = addString(updated, "error|DS.W invalid data");
                                }

                            }
                            else
                            {
                                updated = addString(updated, "error|DS.W invalid data");
                            }
                        }

                    }
                    else
                    {
                        updated = addString(updated, token[j]);
                    }

                }

                if (dsw_found)
                {
                    // create a new array object for each update to the intermediate code
                    newarray = new ArrayList();
                    newarray.Add(updated);
                    newarray.Add(line);
                    intermediateMachineCode[i] = newarray;

                }

            }

        }

        public void processSymbolReferences()
        {

            ArrayList lineTokens;
            ArrayList newarray;
            string[] separator = { ":" };
            string[] token = { };

            string updated = "";
            string line = "";
            bool symbol_found = false;

            string symbolstring = "";
            ushort number = 0;

            newarray = new ArrayList();

            // go through each line
            for (int i = 0; i < intermediateMachineCode.Count; i++)
            {
                // on each line we have a set of instructions
                lineTokens = (ArrayList)intermediateMachineCode[i];

                // split the instructions based on the : character
                token = lineTokens[0].ToString().Split(separator, StringSplitOptions.RemoveEmptyEntries);
                line = lineTokens[1].ToString();

                updated = "";
                symbol_found = false;

                // go through the instructions looking for symbol| , any other entry simply left alone
                for (int j = 0; j < token.Length; j++)
                {
                    if (token[j].StartsWith("symbol|"))
                    {
                        // we have found a symbol reference
                        symbol_found = true;

                        string local = token[j];
                        symbolstring = local.Substring(7, (local.Length - 7));

                        try
                        {
                            number = (ushort)SymbolTable[symbolstring];

                            updated = addString(updated, "actual|$" + supportMethods.FormatUshort(number, Base.hex));
                        }
                        catch (Exception)
                        {
                            updated = addString(updated, "error|Invalid string reference " + symbolstring);
                        }

                    }
                    else
                    {
                        updated = addString(updated, token[j]);
                    }

                }

                if (symbol_found)
                {
                    // create a new array object for each update to the intermediate code
                    newarray = new ArrayList();
                    newarray.Add(updated);
                    newarray.Add(line);
                    intermediateMachineCode[i] = newarray;

                }

            }

        }

        public void processDisplacementSymbolReferences()
        {

            ArrayList lineTokens;
            ArrayList newarray;
            string[] separator = { ":" };
            string[] token = { };

            string updated = "";
            string line = "";
            bool symbol_found = false;

            string symbolstring = "";
            ushort number = 0;

            int currentaddress = 0;

            newarray = new ArrayList();

            // go through each line
            for (int i = 0; i < intermediateMachineCode.Count; i++)
            {
                // on each line we have a set of instructions
                lineTokens = (ArrayList)intermediateMachineCode[i];

                // split the instructions based on the : character
                token = lineTokens[0].ToString().Split(separator, StringSplitOptions.RemoveEmptyEntries);
                line = lineTokens[1].ToString();

                updated = "";
                symbol_found = false;

                // go through the instructions looking for symbol| , any other entry simply left alone
                for (int j = 0; j < token.Length; j++)
                {
                    if (token[j].StartsWith("dis_symbol|"))
                    {
                        currentaddress++;

                        // we have found a symbol reference
                        symbol_found = true;

                        string local = token[j];
                        symbolstring = local.Substring(11, (local.Length - 11));


                        // 'currentaddress' has the address of the NEXT instruction (what would be in the PC)
                        // 'number' is the value returned by looking in the symbol table for the specific symbol

                        try
                        {
                            number = (ushort)SymbolTable[symbolstring];

                            //MessageBox.Show("Address=["
                            //    + supportMethods.FormatUshort((ushort)currentaddress, Base.hex)
                            //    + "] symbol=["
                            //    + supportMethods.FormatUshort((ushort)number, Base.hex)
                            //    + "]");

                            // need to work out the proper value
                            if (number > currentaddress)
                            {
                                // forward reference
                                number = (ushort)(number - currentaddress);
                            }
                            else
                            {
                                // backward reference
                                number = (ushort)(currentaddress - number);

                                // take the 2's complement
                                number = (ushort)(~number);
                                number++;

                            }

                            updated = addString(updated, "actual|$" + supportMethods.FormatUshort(number, Base.hex));
                        }
                        catch (Exception)
                        {
                            updated = addString(updated, "error|Invalid string reference " + symbolstring);
                        }

                    }
                    else if (token[j].StartsWith("actual|"))
                    {
                        currentaddress++;
                        updated = addString(updated, token[j]);
                    }
                    else if (token[j].StartsWith("org|"))
                    {
                        // set the address
                        string fred = token[j].Substring(4, (token[j].Length - 4));
                        ushort newaddress = 0;
                        if (supportMethods.isNumber(fred, ref newaddress) == 1)
                        {
                            currentaddress = newaddress;
                        }
                        else
                        {
                            updated = addString(updated, "error|Incorrect ORG data");
                        }

                        updated = addString(updated, token[j]);

                    }
                    else
                    {
                        // anything else doesn't change the address and we just 
                        // put it back in the list
                        updated = addString(updated, token[j]);
                    }

                }

                if (symbol_found)
                {
                    // create a new array object for each update to the intermediate code
                    newarray = new ArrayList();
                    newarray.Add(updated);
                    newarray.Add(line);
                    intermediateMachineCode[i] = newarray;

                }

            }

        }

        public void processSetSymbolReferences()
        {
            // Count through all our tokens - each one requires a word of storage.
            // update labels accordingly.

            // go through all the intermediate Code

            ArrayList lineTokens;
            string[] separator = { ":" };
            string[] barseparator = { "|" };
            string[] tokens;
            string[] keypair;

            ArrayList newarray;

            ushort address = 0;

            // go through every line
            for (int i = 0; i < intermediateMachineCode.Count; i++)
            {
                lineTokens = (ArrayList)intermediateMachineCode[i];

                bool setsymbolFound = false;

                // go through every tokenset
                if (lineTokens.Count > 0)
                {
                    if (!(lineTokens[0].ToString().Equals("")))
                    {
                        // split lineTokens[0] to get set for that line
                        tokens = lineTokens[0].ToString().Split(separator, StringSplitOptions.None);

                        for (int j = 0; j < tokens.Length; j++)
                        {
                            //MessageBox.Show(tokens[j]);
                            keypair = tokens[j].Split(barseparator, StringSplitOptions.None);

                            switch (keypair[0])
                            {
                                case "frm":
                                case "lpc":
                                case "error":
                                    break;
                                case "org":
                                    supportMethods.isNumber(keypair[1], ref address);
                                    break;
                                case "actual":
                                case "symbol":
                                case "dis_symbol":
                                    address++;
                                    break;
                                case "setsymbol":
                                    //MessageBox.Show("[" + keypair[1].ToString() + "][" + address + "]");
                                    SymbolTable.Remove(keypair[1].ToString());
                                    SymbolTable.Add(keypair[1].ToString(), address);

                                    setsymbolFound = true;

                                    //MessageBox.Show("symbol found on line");

                                    break;
                                default:
                                    //MessageBox.Show("ERROR " + keypair[0]);
                                    break;
                            }

                        }

                        if (setsymbolFound)
                        {
                            // now remove the setsymbol from the intermediateCode
                            // need to update the line
                            newarray = new ArrayList();

                            string newtoken = removeSetSymbolTokens(tokens);

                            newarray.Add(newtoken);
                            newarray.Add(lineTokens[1]);
                            intermediateMachineCode[i] = newarray;

                            setsymbolFound = false;
                        }

                    }
                }
            }


        }

        public void performSave(string filename)
        {
            //MessageBox.Show("Looking to save [" + filename + "]");

            if (filename != null)
            {
                StreamWriter sw;
                string[] details = { };

                try
                {
                    sw = File.CreateText(filename);
                }
                catch (IOException)
                {
                    //MessageBox.Show("File creation failed for some reason, malformed filename?");
                    return;
                }

                sw.Write(finalMachineCode);

                sw.Close();
            }

        }

        public string assemble()
        {
            // initial tokenisation
            initialTokenisation();

            // code all instructions
            initialCoding();

            // get rid of DSWdata
            processDSWReferences();

            // now we need to update all SymbolTable values
            processSetSymbolReferences();
            processSymbolReferences();
            processDisplacementSymbolReferences();

            if (outputTypeMachineCode)
            {
                finalMachineCode = prettyPrintMachineCode();
            }
            else
            {
                finalMachineCode = prettyPrint();
            }

            return (finalMachineCode);
        }



        private ArrayList LoadAssemblerSourceFile(string filename)
        {
            // reads the ASM file

            ArrayList newlist = new ArrayList();

            string contents = "";
            //string error = "";
            string[] tempstring = { };
            string[] lines = { };
            string[] line = { };
            string[] separator = { Environment.NewLine };
            string[] spaceseparator = { " " };
            string[] barseparator = { "|" };
            string[] textseparator = { "\"" };

            if (File.Exists(filename))
            {
                // read the file
                contents = File.ReadAllText(filename);

                lines = contents.Split(separator, StringSplitOptions.None);

                for (int count = 0; count < lines.Length; count++)
                {
                    newlist.Add(lines[count]);
                }
            }
            else
            {
                Console.Error.WriteLine("File [" + filename + "] not found in local or environment locations");
                //error = "    USE " + filename + " - this library is missing";
                //MessageBox.Show(error);
            }

            return (newlist);
        }

        private ArrayList getDCBTokens(String a)
        {
            // here's the rules:
            // -- spaces are allowed inside matching ' chars
            // -- \' is allowed inside matching ' chars
            // -- *  is allowed inside matching ' chars
            // otherwise a space divides tokens and is not stored.

            ArrayList tokens = new ArrayList();

            String nextToken = "";

            char last = '0';
            char current;

            Boolean inText = false;
            Boolean comment = false;

            for (int i = 0; i < a.Length; i++)
            {
                current = a[i];
                if ((current == ' ') && (inText == false))
                {
                    // space char not stored, create new token
                    if (!(nextToken.ToString() == ""))
                    {
                        tokens.Add(nextToken);
                        nextToken = "";
                    }
                }
                else if ((current == ' ') && (inText == true))
                {
                    nextToken = nextToken + current;
                }
                else if ((current == '*') && (inText == false))
                {
                    comment = true;
                }
                else if ((current == '*') && (inText == true))
                {
                    nextToken = nextToken + current;
                }
                else if (current == '\'')
                {
                    // found a '
                    if (last == '\\')
                    {
                        // escaped '
                        nextToken = nextToken + current;
                    }
                    else
                    {
                        if (inText == false)
                        {
                            inText = true;
                        }
                        else
                        {
                            inText = false;
                        }
                        nextToken = nextToken + current;
                    }
                }
                else
                {
                    if (comment == false)
                    {
                        //tokens.Add(current);
                        nextToken = nextToken + current;
                    }
                }

                last = current;
            }

            // add any non-blank token left over
            if (!(nextToken.ToString() == ""))
            {
                tokens.Add(nextToken);
            }

            return (tokens);
        }

        private ArrayList getLineTokens(String line)
        {
            // this process is not to VALIDATE what it finds, it simply
            // turns a given line into a set of tokens to be used by some future validator
            // note that the first token is an  "L" or "-" depending on whether 
            // it starts with a label or not.

            // Final token is always the original contents of the line

            ArrayList tokens = new ArrayList();

            bool comment = false;

            string[] spaceseparator = { " " };
            string[] lineElements;

            // replace tabs with spaces before doing any processing
            String a = line.Replace('\t', ' ');

            string aUpper = a.ToString().ToUpper();

            lineElements = a.Split(spaceseparator, StringSplitOptions.RemoveEmptyEntries);

            if (lineElements.Length > 0)
            {
                if (lineElements[0].StartsWith("*"))
                {
                    // comment - throw away any lines with only a comment on them
                }
                else
                {
                    int startpoint = 0;
                    if (a.StartsWith(lineElements[0]))
                    {
                        // we've found a line beginning with a label so let's flag it with the first token
                        tokens.Add("L"); // label
                        startpoint = 2;
                    }
                    else
                    {
                        tokens.Add("-"); // label
                        startpoint = 1;
                    }

                    comment = false;
                    for (int x = 0; x < lineElements.Length; x++)
                    {

                        if (lineElements[x].StartsWith("*"))
                        {
                            comment = true;
                        }

                        if (comment != true)
                        {
                            tokens.Add(lineElements[x]);
                        }
                    }

                    // DC.B is a special case - becuase cutting it on spaces can mangle the
                    // content - so in this situation we need to backtrack on DC.B instructions to make sure
                    // we properly tokenise the contents of the DC.B parameter

                    // now see if we've eaten a DC.B and need to backtrack
                    if (tokens.Count > startpoint)
                    {
                        if (tokens[startpoint].ToString().ToUpper().Equals("DC.B"))
                        {
                            // it *was* a DC.B - so we need to fix this

                            // firstly, remove the tokens added in error
                            tokens.RemoveRange((startpoint + 1), (tokens.Count - (startpoint + 1)));

                            // now get the contents of the original line AFTER DC.B
                            int index = aUpper.IndexOf("DC.B");
                            string newList = a.Substring(index + 4);

                            ArrayList newList2 = getDCBTokens(newList);

                            for (int ind = 0; ind < newList2.Count; ind++)
                            {
                                tokens.Add(newList2[ind]);
                            }
                        }
                    }
                }
            }

            tokens.Add(line);

            return (tokens);
        }

        private ArrayList normaliseLineTokens(ArrayList a)
        {
            ArrayList newTokens = new ArrayList();

            // let's start having a look at the tokens
            ArrayList lineTokens;

            String label = "";
            String instruction = "";
            String data = "";

            label = "";
            instruction = "";
            data = "";

            lineTokens = (ArrayList)a;

            if (lineTokens.Count > 2)
            {
                // then we have an instruction
                if ((lineTokens[0].ToString() == "L") &&
                    ((lineTokens.Count == 3)
                    || (lineTokens.Count == 4)
                    || (lineTokens.Count == 5)))
                {
                    // right number of tokens for a label line
                    label = lineTokens[1].ToString();
                    if (lineTokens.Count > 3)
                        instruction = lineTokens[2].ToString();
                    if (lineTokens.Count > 4)
                        data = lineTokens[3].ToString();
                    //Text = Text
                    //    + label + "\t"
                    //    + instruction + "\t"
                    //    + data + "\r\n";
                    newTokens.Add(label);
                    newTokens.Add(instruction);
                    newTokens.Add(data);
                    newTokens.Add(lineTokens[lineTokens.Count - 1]);

                }
                else if ((lineTokens[0].ToString() == "-") &&
                    ((lineTokens.Count == 3)
                    || (lineTokens.Count == 4)))
                {
                    // not a label line
                    instruction = lineTokens[1].ToString();
                    if (lineTokens.Count > 3)
                        data = lineTokens[2].ToString();
                    //Text = Text + "\t"
                    //    + instruction + "\t"
                    //    + data + "\r\n";
                    newTokens.Add(label);
                    newTokens.Add(instruction);
                    newTokens.Add(data);
                    newTokens.Add(lineTokens[lineTokens.Count - 1]);

                }
                else
                {
                    newTokens.Add("");
                    newTokens.Add("");
                    newTokens.Add("");
                    newTokens.Add("");
                    newTokens.Add(lineTokens[lineTokens.Count - 1]);

                    //Text = Text
                    //    + "* Error on line [" + lineTokens[lineTokens.Count - 1] + "]\r\n";
                }

            }
            else if (lineTokens.Count > 0)
            {
                // should be no entries with only 2 tokens
                // blank line or comment
                newTokens.Add("");
                newTokens.Add("");
                newTokens.Add("");
                newTokens.Add(lineTokens[lineTokens.Count - 1]);
                //Text = Text + lineTokens[0] + "\r\n";
            }
            else
            {
                // should never occur
            }



            return (newTokens);

        }

        private ArrayList writeInitialTokenSet(ArrayList a)
        {
            // a is an Array of strings - read form the original file
            // output is an ArrayList of ArrayLists of tokens
            ArrayList localTokenSet = new ArrayList();
            ArrayList lineTokens;

            // update the tokenSet
            for (int i = 0; i < a.Count; i++)
            {
                lineTokens = getLineTokens(a[i].ToString());

                localTokenSet.Add(lineTokens);

            }
            return (localTokenSet);
        }

        private ArrayList writeFinalTokenSet(ArrayList a)
        {
            ArrayList newTokens = new ArrayList();

            ArrayList lineTokens;

            for (int i = 0; i < a.Count; i++)
            {
                lineTokens = normaliseLineTokens((ArrayList)a[i]);
                newTokens.Add(lineTokens);
            }

            return (newTokens);
        }

        private ArrayList includeUSE(ArrayList a)
        {
            ArrayList newarray = new ArrayList();
            ArrayList tokens = new ArrayList();

            ArrayList USEarray; // = new ArrayList();

            String token = "";
            String filename = "";

            Boolean found = false;

            for (int i = 0; i < a.Count; i++)
            {
                tokens = getLineTokens(a[i].ToString());

                if (tokens.Count > 1)
                {
                    token = tokens[1].ToString();
                    token = token.ToUpper();
                    if (token == "USE")
                    {
                        //MessageBox.Show("Found a USE");
                        if (tokens.Count > 2)
                        {
                            filename = tokens[2].ToString();
                            //MessageBox.Show("Found a USE [" + filename + "]");
                            found = true;
                        }

                    }

                }

                if (found)
                {
                    // process the file
                    supportMethods.updateFileLocation(ref filename, globalConstants.envName);

                    USEarray = LoadAssemblerSourceFile(filename);

                    if (USEarray.Count > 0)
                    {
                        for (int j = 0; j < USEarray.Count; j++)
                        {
                            newarray.Add(USEarray[j]);
                        }
                    }
                    else
                    {
                        //Console.Error.WriteLine("File [" + filename + "] not found in local or environment locations");
                        //MessageBox.Show("Error - missing library in USE [" + filename + "]");

                    }

                }
                else
                {
                    newarray.Add(a[i]);
                }

                found = false;

            }

            return (newarray);
        }

        private Boolean foundValidUSE(ArrayList a)
        {
            // returns TRUE if at least 1 USE statement
            Boolean retval = false;

            ArrayList tokens = new ArrayList();

            String token = "";
            String filename = "";

            for (int i = 0; i < a.Count; i++)
            {
                tokens = getLineTokens(a[i].ToString());

                if (tokens.Count > 1)
                {
                    token = tokens[1].ToString();
                    token = token.ToUpper();
                    if (token == "USE")
                    {
                        //MessageBox.Show("Found a USE");
                        if (tokens.Count > 2)
                        {
                            filename = tokens[2].ToString();
                            //MessageBox.Show("Found a USE [" + filename + "]");
                            retval = true;
                        }

                    }

                }

            }


            return (retval);
        }

        private ArrayList processEQU(ArrayList a)
        {
            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            if (label.Equals(""))
            {
                // missing label
                newList.Add("error|missing label for EQU");
                newList.Add(line);
            }
            else if (data.Equals(""))
            {
                // missing label
                newList.Add("error|missing data for EQU");
                newList.Add(line);
            }
            else
            {
                ushort num = 0;

                if (supportMethods.isNumber(data, ref num) == 1)
                {
                    //newList.Add("equ_setsymbol|" + label + "|$" + supportMethods.FormatUshort(num, Base.hex));

                    // add the symbol into the symbol list
                    try
                    {
                        SymbolTable.Add(label, num);
                        newList.Add("");
                        newList.Add(line);
                    }
                    catch (Exception)
                    {
                        newList.Add("error|Duplicate symbol in EQU");
                        newList.Add(line);
                    }


                }
                else
                {
                    newList.Add("error|EQU value must be a 16-bit number");
                    newList.Add(line);
                }

            }

            return (newList);
        }

        private ArrayList processBlank(ArrayList a)
        {
            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            newList.Add("");
            newList.Add(line);

            return (newList);
        }

        private ArrayList processSymbol(ArrayList a)
        {

            //Text = Text + "setsymbol|" + label + ":";

            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            newList.Add("setsymbol|" + label);
            newList.Add(line);

            return (newList);
        }

        private ArrayList processDCB(ArrayList a)
        {

            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();
            char c;
            int counter = 0;
            string token = "";
            ushort number = 0;
            string bytestream = "";
            byte newbyte = 0;

            //String newstring = instruction.ToUpper() + " " + data;
            string retstring = "";

            if ((!label.Equals("")))
            {
                retstring = "setsymbol|" + label;
            }

            // now figure out the data and parcel it into words

            //MessageBox.Show("Data = [" + data + "]");

            // go through each character in turn and look for specific things

            ArrayList newarray = new ArrayList();
            newarray = tokeniseDCB(data);

            counter = 0;
            for (int i = 0; i < newarray.Count; i++)
            {
                //MessageBox.Show(i + "  " + newarray[i]);
                token = (string)newarray[i];

                if (token.StartsWith("str|"))
                {
                    // the token is a string
                    token = token.Substring(4, (token.Length - 4));
                    //MessageBox.Show("string [" + token + "]");

                    char preceding = '-';
                    for (int j = 0; j < token.Length; j++)
                    {
                        c = token[j];

                        if (preceding == '\\')
                        {
                            // then this char is a special char '\' , 't' , 'n' , 'r' , '''
                            // any others?

                            if (c == 't')
                            {
                                newbyte = Convert.ToByte('\t');
                            }
                            else if (c == 'n')
                            {
                                newbyte = Convert.ToByte('\n');
                            }
                            else if (c == 'r')
                            {
                                newbyte = Convert.ToByte('\r');
                            }
                            else if (c == '\\')
                            {
                                newbyte = Convert.ToByte('\\');
                            }
                            else if (c == '\'')
                            {
                                newbyte = Convert.ToByte('\'');
                            }
                            number = (ushort)newbyte;
                            counter++;
                            bytestream = bytestream + supportMethods.FormatUshort(number, Base.hex, 2);
                        }
                        else if (c == '\\')
                        {
                            // ignire a backslash as it *should* be a meta char for a special (see above)
                        }
                        else
                        {
                            // just convert it and add it to the bytestream
                            newbyte = Convert.ToByte(c);
                            number = (ushort)newbyte;
                            counter++;
                            bytestream = bytestream + supportMethods.FormatUshort(number, Base.hex, 2);
                        }

                        preceding = c;

                    }

                }
                else
                {
                    // the token is a number
                    token = token.Substring(4, (token.Length - 4));
                    //MessageBox.Show("number [" + token + "]");
                    if (supportMethods.isNumber(token, ref number) == 1)
                    {
                        if (number < 256)
                        {
                            // then it's a byte so encode it
                            counter++;
                            bytestream = bytestream + supportMethods.FormatUshort(number, Base.hex, 2);
                        }
                        else
                        {
                            retstring = addString(retstring, "error|DC.B value not a valid byte");
                        }

                    }
                    else
                    {
                        retstring = addString(retstring, "error|DC.B value not a valid byte");
                    }

                }


            }

            // counter needs to be even, if it isn't we add a null byte on the end of the byte stream then
            // convert into 16-bit actuals.
            if ((counter % 2) != 0)
            {
                bytestream = bytestream + "00";
            }


            //MessageBox.Show(bytestream);

            int count = 1;
            string actuals = "actual|$";
            for (int z = 0; z < bytestream.Length; z++)
            {
                actuals = actuals + bytestream[z];
                if (((count % 4) == 0) && (z != 0) && (z < (bytestream.Length - 4)))
                {
                    actuals = actuals + ":actual|$";
                }
                count++;
            }

            retstring = addString(retstring, actuals);

            newList.Add(retstring);
            newList.Add(line);

            return (newList);
        }

        private ArrayList tokeniseDCB(string s)
        {
            // take the DCB string and separate into array of strings and bytes

            // a bit buggy after a string, I think you could miss a comma and it would still
            // be valid.
            //

            ArrayList retarray = new ArrayList();

            // go through the string and split on commas UNLESS you are currently in a string

            char c;
            char preceding = '-';

            bool inString = false;
            string current = "";

            for (int i = 0; i < s.Length; i++)
            {
                c = s[i];

                current = current + c;

                if ((c == '\'') && (preceding != '\\'))
                {
                    // then we are a string switch
                    if (inString == true)
                    {
                        inString = false;

                        // MessageBox.Show("string [" + current + "]");
                        if (current.Length > 2)
                        {
                            current = current.Substring(1, (current.Length - 2));
                            retarray.Add("str|" + current);
                        }
                        current = "";

                    }
                    else
                    {
                        inString = true;
                    }
                }
                else if ((c == ',') && (inString == false))
                {

                    // if not empty
                    if (current != "")
                    {

                        // the string contains commas - remove them
                        current = current.Replace(",", "");

                        if (current != "")
                        {
                            //MessageBox.Show("NOT string [" + current + "]");
                            retarray.Add("num|" + current);
                        }
                        current = "";

                    }

                }
                else
                {
                    //MessageBox.Show("ELSE NOT string" + current);
                }

                preceding = c;
            }

            // if we have chars left
            if (current != "")
            {
                current = current.Replace(",", "");
                if (current != "")
                {
                    //MessageBox.Show("LAST NOT string [" + current + "]");
                    retarray.Add("num|" + current);
                }
                current = "";
            }

            //MessageBox.Show(s);

            return retarray;
        }

        private ArrayList processLPC(ArrayList a)
        {
            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            String newstring;

            ushort num = 0;

            if (supportMethods.isNumber(data, ref num) == 1)
            {
                //MessageBox.Show("Number " + data + " = " + num);
                newstring = "lpc|" + data;
                newList.Add(newstring);
                newList.Add(line);
            }
            else
            {
                //MessageBox.Show("Error : LPC data not a number");
                newList.Add("error|data is not a 16-bit number");
                newList.Add(line);
            }


            return (newList);
        }

        private ArrayList processDSW(ArrayList a)
        {
            ArrayList newlist = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            String dataString;

            ushort num = 0;

            if (label.Equals(""))
            {
                // no label defined
                dataString = "";
            }
            else
            {
                dataString = "setsymbol|" + label;
            }

            // data MUST be either a valid number or a symbol with a preceding #
            if (data.StartsWith("#"))
            {
                // valid label format (even if later we find out the lable doesn't exist)
                dataString = addString(dataString, "dsw|" + data);
            }
            else if (supportMethods.isNumber(data, ref num) == 1)
            {
                // valid number
                dataString = addString(dataString, "dsw|" + data);
            }
            else
            {
                dataString = addString(dataString, "error|DS.W data bad format");
            }

            newlist.Add(dataString);
            newlist.Add(line);

            return (newlist);
        }

        private ArrayList processDCW(ArrayList a)
        {
            // can have an optional label
            // data can be multiple comma-separated entries
            // data can be numeric or #symbol

            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            String datastring = "";
            string[] token = { };
            string[] separator = { "," };

            string value = "";

            ushort num = 0;

            // deal with label first
            if (label != "")
            {
                datastring = "setsymbol|" + label;
            }

            // now split the data by commas and parse each in turn
            token = data.Split(separator, StringSplitOptions.RemoveEmptyEntries);

            for (int i = 0; i < token.Length; i++)
            {
                value = token[i];

                if (supportMethods.isNumber(value, ref num) == 1)
                {
                    datastring = addString(datastring, "actual|$" +
                        (supportMethods.FormatUshort((ushort)num, Base.hex).ToUpper()));

                }
                else
                {
                    // if it has a leading # it's a symbol reference otherwise it's an error
                    if (value.StartsWith("#"))
                    {
                        datastring = addString(datastring, "symbol|" + value.Substring(1, (value.Length - 1)));
                    }
                    else
                    {
                        datastring = addString(datastring, "error|data is not a 16-bit number");
                    }


                    //newList.Add("error|data is not a 16-bit number");
                    //newList.Add(line);
                }

            }
            newList.Add(datastring);
            newList.Add(line);

            return (newList);
        }

        private ArrayList processORG(ArrayList a)
        {
            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            String newstring;

            ushort num = 0;

            if (label != "")
            {
                // labels not allowed
                newList.Add("error|label not allowed");
                newList.Add(line);
            }
            else
            {
                if (supportMethods.isNumber(data, ref num) == 1)
                {
                    //MessageBox.Show("Number " + data + " = " + num);
                    newstring = "org|" + data;
                    newList.Add(newstring);
                    newList.Add(line);
                }
                else
                {
                    newList.Add("error|data is not a 16-bit number");
                    newList.Add(line);
                }
            }



            return (newList);
        }

        private ArrayList processInstruction(ArrayList a)
        {

            ArrayList newList = new ArrayList();

            String label = a[0].ToString();
            String instruction = a[1].ToString();
            String data = a[2].ToString();
            String line = a[3].ToString();

            string textstring = "";

            int actualCount = 0;

            // deal with any labels
            if ((!label.Equals("")))
            {
                textstring = textstring + "setsymbol|" + label + ":";
            }

            // now, we have a line of assembler to decode

            string assembler = instruction + " " + data;

            string[] srcTokens = tokeniseAssembler(assembler);

            // look at individual tokens and label registers, numbers and strings
            srcTokens = updateSrc(srcTokens);

            // lets make it easy to print
            //string stringAssembler = stringifyAssembler(srcTokens);

            bool gotmatch = false;
            for (int j = 0; j < 256; j++)
            {
                // tokenise each firmware instruction and check for a match
                string frmMnemonic = firmware.getMnemonic((ushort)j);

                string[] frmTokens = tokeniseAssembler(frmMnemonic);

                // now perform the match
                ushort opcode = 0;
                if (assemblerMatch(frmTokens, srcTokens, j, ref opcode) == true)
                {

                    gotmatch = true;

                    //string stringFirmware = stringifyAssembler(frmTokens);

                    // don't bother trying to match any more
                    textstring = textstring + "actual|$" + supportMethods.FormatUshort(opcode, Base.hex);
                    actualCount++;

                    // now we can include any numbers and strings
                    for (int tokens = 0; tokens < srcTokens.Length; tokens++)
                    {
                        if (srcTokens[tokens].StartsWith("actual|"))
                        {
                            textstring = textstring + ":" + srcTokens[tokens];
                            actualCount++;
                        }
                        else if (srcTokens[tokens].StartsWith("symbol|"))
                        {
                            actualCount++;

                            // either its a label or a displacement label
                            if (string.Equals(frmTokens[tokens], "dis", StringComparison.CurrentCultureIgnoreCase))
                            {
                                textstring = textstring + ":dis_" + srcTokens[tokens];
                            }
                            else
                            {
                                textstring = textstring + ":" + srcTokens[tokens];
                            }
                        }
                    }

                    if ((actualCount % 2) == 1)
                    {
                        // we have an odd number of words, so pad out
                        textstring = textstring + ":actual|$0000";
                    }
                    actualCount = 0;

                    newList.Add(textstring);
                    newList.Add(line);
                    //Text = "";

                    break;
                }
            }

            // if we got to here and we still don't have a match then we need to report a bad instruction
            if (gotmatch == false)
            {
                newList.Add("error|unknown instruction");
                newList.Add(line);
            }

            return (newList);
        }

        private ArrayList processError(ArrayList a)
        {
            ArrayList newList = new ArrayList();
            String line = a[4].ToString();

            newList.Add("error|wrong number of tokens on line");
            newList.Add(line);

            return (newList);
        }

        private ArrayList processBadInstruction(ArrayList a)
        {
            ArrayList newList = new ArrayList();
            String line = a[3].ToString();

            newList.Add("error|unknown instruction");
            newList.Add(line);

            return (newList);
        }

        private String removeSetSymbolTokens(string[] s)
        {
            string retval = "";

            string[] separator = { "|" };
            string[] tokens;

            //int regcount = 0;

            //int regkey = 0;

            //ushort currentOpcode = 0;
            string retstring = "";

            // We'll have an opcode reference followed by zero, one or two register references
            // other tokens should be left as they are

            foreach (string tokenSet in s)
            {
                //MessageBox.Show("Handed [" + tokenSet + "]");

                // split each one into a key and a value
                tokens = tokenSet.Split(separator, StringSplitOptions.None);

                //MessageBox.Show(token);

                if (tokens[0].Equals("setsymbol"))
                {
                    //MessageBox.Show("WAS a setsymbol");
                }
                else
                {
                    //MessageBox.Show("NOT a setsymbol");
                    if (retstring.Equals(""))
                    {
                        retstring = tokens[0] + "|" + tokens[1];
                    }
                    else
                    {
                        retstring = retstring + ":" + tokens[0] + "|" + tokens[1];
                    }
                }

            }

            retval = retstring;


            return (retval);
        }

        private bool assemblerMatch(string[] frm, string[] src, int index, ref ushort opcode)
        {
            // returns true if a match is found, opcode is updated with the opcode and any register
            // references included.

            bool retval = true;
            int registerReferences = 0;
            ushort regIndex = 0;
            ushort regStore = 0;

            // do the match
            if (frm.Length != src.Length)
            {
                // if not the same number of tokens then definitely no match
                retval = false;
            }
            else
            {
                // look at the entries in the src and figure out which are symbols and which are
                // numbers
                //src = updateSrc(src);


                // same number of tokens so go through them
                for (int i = 0; i < frm.Length; i++)
                {
                    if (string.Equals(frm[i], src[i], StringComparison.CurrentCultureIgnoreCase))
                    {
                        // We have an exact match
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                    }
                    else if ((string.Equals(frm[i], "addr", StringComparison.CurrentCultureIgnoreCase))
                        && (src[i].StartsWith("actual|", StringComparison.CurrentCultureIgnoreCase))
                        )
                    {
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                    }
                    else if ((string.Equals(frm[i], "addr", StringComparison.CurrentCultureIgnoreCase))
                        && (src[i].StartsWith("symbol|", StringComparison.CurrentCultureIgnoreCase))
                        )
                    {
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                    }
                    else if ((string.Equals(frm[i], "data", StringComparison.CurrentCultureIgnoreCase))
                        && (src[i].StartsWith("symbol|", StringComparison.CurrentCultureIgnoreCase))
                        )
                    {
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                    }
                    else if ((string.Equals(frm[i], "data", StringComparison.CurrentCultureIgnoreCase))
                        && (src[i].StartsWith("actual|", StringComparison.CurrentCultureIgnoreCase))
                        )
                    {
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                    }
                    else if ((string.Equals(frm[i], "dis", StringComparison.CurrentCultureIgnoreCase))
                        && (src[i].StartsWith("symbol|", StringComparison.CurrentCultureIgnoreCase))
                        )
                    {
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                    }
                    else if ((string.Equals(frm[i], "dis", StringComparison.CurrentCultureIgnoreCase))
                        && (src[i].StartsWith("actual|", StringComparison.CurrentCultureIgnoreCase))
                        )
                    {
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                    }
                    else if ((string.Equals(frm[i], "R", StringComparison.CurrentCultureIgnoreCase))
                        && (src[i].StartsWith("register|", StringComparison.CurrentCultureIgnoreCase))
                        )
                    {
                        //Text = Text + "M: [" + frm[i] + "][" + src[i] + "]\r\n";
                        if (registerReferences == 0)
                        {
                            registerReferences++;
                            regIndex = getRegister(src[i]);
                            regStore = (ushort)(regIndex * 16);

                        }
                        else if (registerReferences == 1)
                        {
                            registerReferences++;
                            regIndex = getRegister(src[i]);
                            regStore = (ushort)(regStore + regIndex);
                        }
                        else
                        {
                            // more than 2 register references, any others ignored
                        }

                    }
                    else
                    {
                        retval = false;
                        break;
                    }

                }
            }

            if (retval == true)
            {
                // figure out the returning ushort opcode

                opcode = (ushort)(index * 256);

                // now add in any register references
                opcode = (ushort)(opcode + regStore);
            }
            else
            {
                // otherwise just zeroise it
                opcode = 0;
            }

            return retval;
        }

        private ushort getRegister(string reg)
        {
            ushort retval = 0;

            string[] splitter = { "|" };
            string[] regTokens = reg.Split(splitter, StringSplitOptions.None);
            //0-7	d0-d7
            //8	    pc
            //9	    mar
            //10	mdr
            //11	alux
            //12	aluy
            //13	alur
            //14	usp
            //15	ssp

            if (regTokens.Length != 0)
            {
                regTokens[1] = regTokens[1].ToLower();
                switch (regTokens[1])
                {
                    case "d0":
                        retval = 0;
                        break;
                    case "d1":
                        retval = 1;
                        break;
                    case "d2":
                        retval = 2;
                        break;
                    case "d3":
                        retval = 3;
                        break;
                    case "d4":
                        retval = 4;
                        break;
                    case "d5":
                        retval = 5;
                        break;
                    case "d6":
                        retval = 6;
                        break;
                    case "d7":
                        retval = 7;
                        break;
                    case "usp":
                        retval = 8;
                        break;
                    case "ssp":
                        retval = 9;
                        break;
                }
            }

            return retval;
        }

        private string[] updateSrc(string[] src)
        {
            string[] updatedSrc = src;

            if (src.Length > 0)
            {
                for (int i = 1; i < src.Length; i++)
                {
                    if (string.Equals(src[i], "(", StringComparison.CurrentCultureIgnoreCase))
                    {
                    }
                    else if (string.Equals(src[i], ")", StringComparison.CurrentCultureIgnoreCase))
                    {
                    }
                    else if (string.Equals(src[i], ",", StringComparison.CurrentCultureIgnoreCase))
                    {
                    }
                    else if (string.Equals(src[i], "+", StringComparison.CurrentCultureIgnoreCase))
                    {
                    }
                    else if (string.Equals(src[i], "+", StringComparison.CurrentCultureIgnoreCase))
                    {
                    }
                    else if (string.Equals(src[i], "#", StringComparison.CurrentCultureIgnoreCase))
                    {
                    }
                    else
                    {
                        // it's either a number or a label or a register
                        //0-7	d0-d7
                        //8	    pc
                        //9	    mar
                        //10	mdr
                        //11	alux
                        //12	aluy
                        //13	alur
                        //14	usp
                        //15	ssp

                        if (
                               (string.Equals(src[i], "D0", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "D1", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "D2", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "D3", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "D4", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "D5", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "D6", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "D7", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "pc", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "mar", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "mdr", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "alux", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "aluy", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "alur", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "usp", StringComparison.CurrentCultureIgnoreCase))
                            || (string.Equals(src[i], "ssp", StringComparison.CurrentCultureIgnoreCase))
                            )
                        {
                            updatedSrc[i] = "register|" + src[i];
                        }
                        else
                        {
                            ushort result = 0;
                            if (supportMethods.isNumber(src[i], ref result) == 1)
                            {
                                // it is a valid number
                                updatedSrc[i] = "actual|$" + supportMethods.FormatUshort(result, Base.hex);
                            }
                            else
                            {
                                // it is a string
                                updatedSrc[i] = "symbol|" + src[i];
                            }
                        }
                    }

                }
            }


            return (updatedSrc);
        }

        private string[] tokeniseAssembler(string a)
        {
            // this function extracts the individual elements of an assembler string
            // that consists of an instruction and data.  It can come from a source file 
            // or a firmware file.
            // For example receiving [MOVE (D0),$1001] should result in [MOVE][(][D0][)][,][$1001]
            // and [MOVE (R),addr] should result in [MOVE][(][R][)][,][addr]
            // you can see that direct comparisons should then be possible

            string[] result = { };

            // todo

            string[] barseparator = { "|" };

            string tempdata = a;
            string[] dataTokens;

            tempdata = tempdata.Replace(",", "|,|");
            tempdata = tempdata.Replace("(", "|(|");
            tempdata = tempdata.Replace(")", "|)|");
            tempdata = tempdata.Replace("+", "|+|");
            tempdata = tempdata.Replace("-", "|-|");
            tempdata = tempdata.Replace(" ", "||");
            tempdata = tempdata.Replace("#", "|#|");
            tempdata = tempdata.Replace("\t", "||"); // haven't tested if tabs are removed

            dataTokens = tempdata.Split(barseparator, StringSplitOptions.RemoveEmptyEntries);

            result = dataTokens;

            return (result);
        }

        private string addString(string current, string newstring)
        {
            if (current.Equals(""))
            {
                current = newstring;
            }
            else
            {
                current = current + ":" + newstring;
            }

            return current;
        }






    }
}
