//Copyright (c) 2009, Mark S. Burrell
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are 
//permitted provided that the following conditions are met:

//    * Redistributions of source code must retain the above copyright notice, this list 
//of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this 
//list of conditions and the following disclaimer in the documentation and/or other 
//materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
//EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
//OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
//SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
//OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.Collections.Generic;

namespace Sedici.Hardware
{
    public class Chip_Clock : Chip
    {
        // Constants
        private const ushort SIZE = 8;

        private const ushort SEC = 0;
        private const ushort MIN = 1;
        private const ushort HR = 2;
        private const ushort DAY = 3;
        private const ushort MON = 4;
        private const ushort YR = 5;
        private const ushort TIMER = 6;

        private ushort[] memory = new ushort[SIZE];

        private static ushort internalTimer = 0;

        public Chip_Clock(ushort value)
        {
            BaseAddress = value;
            HardReset();
            return;
        }

        override public ushort Size
        {
            get
            {
                return (SIZE);
            }
        }

        override public ushort Read(ushort address)
        {
            DateTime datetime;
            ushort value = 0;

            // convert the memory address to the true card address
            address = (ushort)(address - BaseAddress);

            // get the current date/time attribute
            datetime = DateTime.Now;
            switch (address)
            {
                case SEC:
                    value = (ushort)datetime.Second;
                    break;
                case MIN:
                    value = (ushort)datetime.Minute;
                    break;
                case HR:
                    value = (ushort)datetime.Hour;
                    break;
                case DAY:
                    value = (ushort)datetime.Day;
                    break;
                case MON:
                    value = (ushort)datetime.Month;
                    break;
                case YR:
                    value = (ushort)datetime.Year;
                    break;
                case TIMER:
                    value = memory[address];
                    break;
                default:
                    // unused locations - do nothing
                    break;
            }
            return (value);
        }

        override public void Write(ushort address, ushort value)
        {
            // convert the memory address to the true card address
            address = (ushort)(address - BaseAddress);

            switch (address)
            {
                case TIMER:
                    // only value between 0 - 255 allowed
                    value = (ushort)(value & 0x00FF);
                    memory[address] = value;
                    internalTimer = 0;
                    break;
                default:
                    // unused locations - do nothing
                    break;
            }
        }

        override public void HardReset()
        {
            // Initialise memory
            for (int i = 0; i < SIZE; i++)
            {
                memory[i] = 0x0000;
            }
            internalTimer = 0;
            return;
        }

        override public void SoftReset()
        {
            // synonymous with a hard reset
            HardReset();
            return;
        }

        public bool InterruptRaised(ref ushort vector)
        {
            bool interruptRaised = false;
            DateTime datetime;

            ushort min = 0;
            ushort sec = 0;

            ushort currentTime = 0;

            if (memory[TIMER] != 0)
            {
                // we need to check to see if we should fire an interrupt 6
                datetime = DateTime.Now;
                min = (ushort)datetime.Minute;
                sec = (ushort)datetime.Second;

                currentTime = (ushort)((((min * 60) + sec) % 360) + 1); // range 1 - 360

                if (internalTimer == currentTime)
                {
                    // set the interrupt vector
                    vector = 6;

                    // trigger the interrupt
                    interruptRaised = true;

                    // set the new Timer time
                    internalTimer = (ushort)(((currentTime + memory[TIMER] - 1) % 360) + 1);
                }
                else if (internalTimer == 0)
                {
                    // set the new Timer time
                    internalTimer = (ushort)(((currentTime + memory[TIMER] - 1) % 360) + 1);
                }
            }
            return(interruptRaised);
        }
    }
}