//Copyright (c) 2009, Mark S. Burrell
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are 
//permitted provided that the following conditions are met:

//    * Redistributions of source code must retain the above copyright notice, this list 
//of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this 
//list of conditions and the following disclaimer in the documentation and/or other 
//materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
//EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
//OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
//SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
//OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.Collections.Generic;

namespace Sedici.Hardware
{
    public class Chip_IO : Chip
    {

        // define our own private semaphore class
        private class semaphore
        {
            private Random randomseed = new Random();

            private int seed;

            public void Seed()
            {
                //set the seed to a random value between 3 - 8
                seed = randomseed.Next(2, 5);
            }

            public bool Release()
            {
                if (seed == 0)
                {
                    return (true);
                }
                else
                {
                    seed--;
                }
                return (false);
            }
        }


        // Constants
        private const ushort SIZE = 4;

        public const ushort IDR = 0;
        public const ushort ISR = 1;
        public const ushort ODR = 2;
        public const ushort OSR = 3;

        private ushort[] memory = new ushort[SIZE];

        private bool charinIDR = false;
        private ushort IDRchar = 0;
        private ushort ODRchar = 0;
        private string outputbuffer = "";

        private semaphore InputSemaphore = new semaphore();
        private semaphore OutputSemaphore = new semaphore();

        public Chip_IO(ushort value)
        {
            BaseAddress = value;
            HardReset();
            return;
        }

        override public ushort Size
        {
            get
            {
                return (SIZE);
            }
        }

        override public ushort Read(ushort address)
        {
            ushort value = 0;

            // convert the memory address to the true card address
            address = (ushort)(address - BaseAddress);

            switch (address)
            {
                case IDR:
                    value = memory[address];
                    memory[ISR] = 0; // signal key used
                    break;
                case ISR:
                    if (charinIDR)
                    {
                        if (InputSemaphore.Release())
                        {
                            memory[ISR] = 1;
                            charinIDR = false;
                            // delayed write into the IDR
                            memory[IDR] = IDRchar;
                        }
                    }
                    value = memory[address];
                    break;
                case ODR:
                    value = memory[address];
                    break;
                case OSR:
                    // we are checking the OSR to see if we can write to the ODR
                    value = memory[address];
                    if (value == 0)
                    {
                        if (OutputSemaphore.Release())
                        {
                            value = 1;
                            memory[OSR] = 1;
                        }
                    }
                    break;
                default:
                    // unused locations - do nothing
                    break;
            }
            return (value);
        }

        override public void Write(ushort address, ushort value)
        {
            // convert the memory address to the true card address
            address = (ushort)(address - BaseAddress);

            switch (address)
            {
                case IDR:
                    charinIDR = true;
                    InputSemaphore.Seed();
                    value = (ushort)(value & 0x007F);                    
                    // don't directly do the IDR write, instead set the semaphore
                    // and do the write when it releases
                    IDRchar = value;
                    break;
                case ISR:
                    // can't update the ISR so don't bother writing
                    break;
                case ODR:
                    value = (ushort)(value & 0x007F);

                    memory[address] = value;
                    OutputSemaphore.Seed();

                    if (memory[OSR] == 1)
                    {
                        memory[OSR] = 0;
                        ODRchar = memory[ODR];
                        outputbuffer = outputbuffer + (char)memory[address];
                    }
                    else
                    {
                        // commented this out - can't remember what it was for? prior to outputbuffer???
                        //write the last char
                        //     Console.Write((char)ODRchar); 
                    }
                    break;
                case OSR:
                    // can't update the OSR so don't bother writing
                    break;
                default:
                    // unused locations - do nothing
                    break;
            }
        }

        override public void HardReset()
        {
            // Initialise memory
            for (int i = 0; i < SIZE; i++)
            {
                memory[i] = 0x0000;
            }
            memory[OSR] = 1;
            memory[ISR] = 0;
            
            OutputSemaphore.Seed();
            InputSemaphore.Seed();
            return;
        }

        override public void SoftReset()
        {
            // synonymous with a hard reset
            HardReset();
            return;
        }

        public string OutputBuffer()
        {
            string s = outputbuffer;
            outputbuffer = "";
            return (s);
        }
    }
}