//Copyright (c) 2009, Mark S. Burrell
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are 
//permitted provided that the following conditions are met:

//    * Redistributions of source code must retain the above copyright notice, this list 
//of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this 
//list of conditions and the following disclaimer in the documentation and/or other 
//materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
//EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
//OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
//SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
//OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Sedici.Hardware
{
    // Order in here is important - it indicates the index for r0 and r1.
    // And yes, I know CU isn't a register really
    public enum CPURegister { D0,   D1,   D2,   D3,   D4,   D5,  D6,  D7, 
                              USP, SSP,   PC,  MAR,  MDR, ALUx, ALUy, ALUr, 
                              IRoperand,  IRcmd,  INC,   CU,  PSR }

    public enum PSRflag { V, N, Z, C, I, E, S }

    public enum Base { binary, hex }

    public class CPU
    {

        private const string CPU_VER = "0.1.0";

        // bit masks
        private const ushort maskS = 0x8000;
        private const ushort maskI = 0x2000;
        private const ushort maskE = 0x1000;
        private const ushort maskV = 0x0008;
        private const ushort maskN = 0x0004;
        private const ushort maskZ = 0x0002;
        private const ushort maskC = 0x0001;

        private const ushort maskv3 = 0x0400;  // vxx
        private const ushort maskv2 = 0x0200;  // xvx
        private const ushort maskv1 = 0x0100;  // xxv

        // positions in the opcode arrays
        private const ushort FETCH = 256;
        private const ushort INTERRUPT = 257;

        // define Registers
        private const int REGISTERSIZE = 21;
        private ushort[] Register = new ushort[REGISTERSIZE]; 

        // define flags
        private const int FLAGSIZE = 7;
        private ushort[] Flag = new ushort[FLAGSIZE]; 

        private ulong machinecycles = 0;

        // define Opcodes object
        public Firmware firmware = new Firmware();

        // what the heck is this doing in HERE?????? Should be part of console?
        private ushort interruptBase = 0x0030;

        private bool running;

        private Messages messageList;

        public CPU(ref Messages m)
        {
            messageList = m;
        }

        public void LoadFirmware(string filename)
        {
            firmware.LoadFirmware(filename);
            return;
        }

        public void LoadFirmware2(string filename)
        {
            firmware.LoadFirmwareOldFormat(filename);
            return;
        }

        public int RunSingleMicroInstruction(string microinstruction, ref Memory mem)
        {
            // count each microinstruction as taking 1 clock cycle
            machinecycles++;

            // variable for ushort overflow checks
            int integerResult = 0;

            // used in SUB
            ushort partialResult = 0;

            // return value
            int RetVal = 0;

            int r = 0;
            string error = "";

            // we do this because all micro-instructions can be written in the
            // instruction set case-insensitive
            string lower_microinstruction = microinstruction.ToLower();

            switch (lower_microinstruction)
            {
                case "illegal":
                    error = "Illegal microinstruction encountered";
                    //MessageStore.AddMessage(messagesType.Error, 0, error);
                    //Console.WriteLine(error);
                    messageList.AddError(error);

                    running = false;
                    break;

                // The PSR Flag Tests
                case "if(psr(c)==1)":
                    if ((Register[(int)CPURegister.PSR] & maskC) == 0) RetVal = 1;
                    break;

                case "if(psr(c)==0)":
                    if ((Register[(int)CPURegister.PSR] & maskC) == maskC) RetVal = 1;
                    break;

                case "if(psr(n)==1)":
                    if ((Register[(int)CPURegister.PSR] & maskN) == 0) RetVal = 1;
                    break;

                case "if(psr(n)==0)":
                    if ((Register[(int)CPURegister.PSR] & maskN) == maskN) RetVal = 1;
                    break;

                case "if(psr(z)==1)":
                    if ((Register[(int)CPURegister.PSR] & maskZ) == 0) RetVal = 1;
                    break;

                case "if(psr(z)==0)":
                    if ((Register[(int)CPURegister.PSR] & maskZ) == maskZ) RetVal = 1;
                    break;

                case "if(psr(v)==1)":
                    if ((Register[(int)CPURegister.PSR] & maskV) == 0) RetVal = 1;
                    break;

                case "if(psr(v)==0)":
                    if ((Register[(int)CPURegister.PSR] & maskV) == maskV) RetVal = 1;
                    break;


                // The ALU Operations
                case "alur=[alux]+[aluy]":
                    integerResult = Register[(int)CPURegister.ALUx] + Register[(int)CPURegister.ALUy];
                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    UpdateFlagC(integerResult);
                    UpdateFlagZ();
                    UpdateFlagN();

                    if ((((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000) 
                        && ((Register[(int)CPURegister.ALUy] & 0x8000) == 0x8000) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0)) 
                        ||
                        (((Register[(int)CPURegister.ALUx] & 0x8000) == 0) 
                        && ((Register[(int)CPURegister.ALUy] & 0x8000) == 0) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0x8000)))
                        SetFlag(PSRflag.V, 1);
                    else
                        SetFlag(PSRflag.V, 0);

                    break;

                case "alur=[alux]+[aluy]+[psr(c)]":
                    integerResult = Register[(int)CPURegister.PSR] & maskC;
                    if (integerResult == maskC) integerResult = 1; else integerResult = 0;
                    integerResult = Register[(int)CPURegister.ALUx] + Register[(int)CPURegister.ALUy] + integerResult;
                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    UpdateFlagC(integerResult);
                    UpdateFlagZ();
                    UpdateFlagN();

                    if ((((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000) 
                        && ((Register[(int)CPURegister.ALUy] & 0x8000) == 0x8000) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0)) 
                        ||
                        (((Register[(int)CPURegister.ALUx] & 0x8000) == 0) 
                        && ((Register[(int)CPURegister.ALUy] & 0x8000) == 0) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0x8000)))
                        SetFlag(PSRflag.V, 1);
                    else
                        SetFlag(PSRflag.V, 0);

                    break;

                case "alur=[alux]-[aluy]":
                    //Really adding the two's complement of ALUy
                    partialResult = (ushort)(~(Register[(int)CPURegister.ALUy]) + 1);
                    integerResult = Register[(int)CPURegister.ALUx] + ~(Register[(int)CPURegister.ALUy]) + 1;

                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    UpdateFlagC(integerResult);
                    UpdateFlagZ();
                    UpdateFlagN();

                    if ((((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000) 
                        && ((partialResult & 0x8000) == 0x8000) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0)) 
                        ||
                        (((Register[(int)CPURegister.ALUx] & 0x8000) == 0) 
                        && ((partialResult & 0x8000) == 0) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0x8000)))
                        SetFlag(PSRflag.V, 1);
                    else
                        SetFlag(PSRflag.V, 0);
                    break;

                case "alur=[alux]<<<1":
                    // logical shift left
                    integerResult = Register[(int)CPURegister.ALUx] * 2;
                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    if ((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000)
                        SetFlag(PSRflag.C, 1);
                    else
                        SetFlag(PSRflag.C, 0);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);
                    break;

                case "alur=[alux]<<1":
                    // arithmetic shift left
                    integerResult = Register[(int)CPURegister.ALUx] * 2;
                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    if ((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000)
                        SetFlag(PSRflag.C, 1);
                    else
                        SetFlag(PSRflag.C, 0);

                    UpdateFlagZ();
                    UpdateFlagN();

                    if (((Register[(int)CPURegister.ALUx] & 0x8000) == 0)
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0x8000))
                        SetFlag(PSRflag.V, 1);
                    else
                        SetFlag(PSRflag.V, 0);

                    break;

                case "alur=[alux]>>>1":
                    // logical shift right
                    Register[(int)CPURegister.ALUr] =
                        (ushort)(Register[(int)CPURegister.ALUx] / 2);

                    if ((Register[(int)CPURegister.ALUx] & 0x0001) == 0x0001)
                        SetFlag(PSRflag.C, 1);
                    else
                        SetFlag(PSRflag.C, 0);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);
                    break;

                case "alur=[alux]>>1":
                    // arithmetic shift right
                    Register[(int)CPURegister.ALUr] =
                        (ushort)(Register[(int)CPURegister.ALUx] / 2);
                    if ((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000)
                    {
                        // if the MSB of ALUx is 1 then we need to make sure the MSB of
                        // ALUr is too
                        Register[(int)CPURegister.ALUr] =
                            (ushort)(Register[(int)CPURegister.ALUr] + 0x8000);
                    }

                    if ((Register[(int)CPURegister.ALUx] & 0x0001) == 0x0001)
                        SetFlag(PSRflag.C, 1);
                    else
                        SetFlag(PSRflag.C, 0);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);
                    break;

                case "alur=[alux]<<=1":
                    // Rotate left
                    //
                    // TODO
                    //
                    integerResult = Register[(int)CPURegister.ALUx] * 2;
                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    if ((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000)
                    {
                        SetFlag(PSRflag.C, 1);
                        Register[(int)CPURegister.ALUr] =
                            (ushort)(Register[(int)CPURegister.ALUr] + 0x0001);
                    }
                    else
                        SetFlag(PSRflag.C, 0);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);
                    break;

                case "alur=[alux]=>>1":
                    // Rotate right
                    Register[(int)CPURegister.ALUr] =
                        (ushort)(Register[(int)CPURegister.ALUx] / 2);

                    if ((Register[(int)CPURegister.ALUx] & 0x0001) == 0x0001)
                    {
                        SetFlag(PSRflag.C, 1);
                        Register[(int)CPURegister.ALUr] =
                            (ushort)(Register[(int)CPURegister.ALUr] + 0x8000);
                    }
                    else
                        SetFlag(PSRflag.C, 0);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);

                    break;

                case "alur=[alux]&[aluy]":
                    Register[(int)CPURegister.ALUr] = 
                        (ushort)(Register[(int)CPURegister.ALUx] & Register[(int)CPURegister.ALUy]);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);
                    SetFlag(PSRflag.C, 0);
                    break;


                case "alur=[alux]|[aluy]":
                    Register[(int)CPURegister.ALUr] 
                        = (ushort)(Register[(int)CPURegister.ALUx] | Register[(int)CPURegister.ALUy]);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);
                    SetFlag(PSRflag.C, 0);
                    break;

                case "alur=~[alux]":
                    Register[(int)CPURegister.ALUr] = (ushort)(~Register[(int)CPURegister.ALUx]);

                    UpdateFlagZ();
                    UpdateFlagN();
                    SetFlag(PSRflag.V, 0);
                    SetFlag(PSRflag.C, 0);
                    break;

                case "alur=[alux]+1":
                    integerResult = Register[(int)CPURegister.ALUx] + 1;
                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    UpdateFlagC(integerResult);
                    UpdateFlagZ();
                    UpdateFlagN();

                    if (((Register[(int)CPURegister.ALUx] & 0x8000) == 0) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0x8000))
                        SetFlag(PSRflag.V, 1);
                    else
                        SetFlag(PSRflag.V, 0);

                    break;

                case "alur=[alux]-1":
                    integerResult = Register[(int)CPURegister.ALUx] + 0xFFFF;
                    Register[(int)CPURegister.ALUr] = (ushort)integerResult;

                    UpdateFlagC(integerResult);
                    UpdateFlagZ();
                    UpdateFlagN();

                    if (((Register[(int)CPURegister.ALUx] & 0x8000) == 0x8000) 
                        && ((Register[(int)CPURegister.ALUr] & 0x8000) == 0))
                        SetFlag(PSRflag.V, 1);
                    else
                        SetFlag(PSRflag.V, 0);

                    break;

                // Misc Processor Operations
                case "nop":
                    // do nothing
                    break;

                case "halt":
                    running = false;
                    break;

                case "psr(i)=1":
                    SetFlag(PSRflag.I, 1);
                    break;

                case "psr(i)=0":
                    SetFlag(PSRflag.I, 0);
                    break;

                case "psr(e)=1":
                    SetFlag(PSRflag.E, 1);
                    break;

                case "psr(e)=0":
                    SetFlag(PSRflag.E, 0);
                    break;

                case "psr(s)=1":
                    SetFlag(PSRflag.S, 1);
                    break;

                case "psr(s)=0":
                    SetFlag(PSRflag.S, 0);
                    break;

                // The Data Movements

                case "cu<-[ir(opcode)]":
                    Register[(int)CPURegister.CU] = (ushort)(Register[(int)CPURegister.IRcmd] / 256); // 8-bit transfer
                    break;

                case "m[mar]<-[mdr]":
                    mem.Write(Register[(int)CPURegister.MAR], Register[(int)CPURegister.MDR]);
                    break;

                case "mdr<-[m[mar]]":
                    Register[(int)CPURegister.MDR] = mem.Read(Register[(int)CPURegister.MAR]);
                    break;

                case "pc<-[inc]":
                    Register[(int)CPURegister.PC] = Register[(int)CPURegister.INC];
                    break;

                case "inc<-[pc]":
                    Register[(int)CPURegister.INC] = (ushort)(Register[(int)CPURegister.PC] + 1);
                    break;


                // it could only be a generic data movement
                default:
                    // The first thing to do is make sure we have something in the right format
                    // and extract SOURCE and DEST

                    //Console.WriteLine("Generic data movement encountered");

                    string[] move = { "<-" };
                    string[] leftbracket = { "[" };
                    string[] rightbracket = { "]" };

                    string[] parts = { };

                    string src = "";
                    string dest = "";

                    parts = lower_microinstruction.Split(move, StringSplitOptions.None);

                    if (parts.Length == 2)
                    {
                        // we have the right number of params
                        dest = parts[0];

                        // now let's see if we can remove the brackets of the second param
                        parts = parts[1].Split(leftbracket, StringSplitOptions.None);

                        // we should have 2 parts again
                        if (parts.Length == 2)
                        {
                            parts = parts[1].Split(rightbracket, StringSplitOptions.None);

                            // again 2 parts left hopefully
                            if (parts.Length == 2)
                            {
                                // first one should be the source
                                src = parts[0];
                            }
                        }
                    }

                    //Console.WriteLine("SRC = [" + src + "]  DEST = [" + dest + "]");

                    // Now check if the SOURCE|DEST pair are both in the 'approved' list
                    bool src_illegal = false;
                    bool dest_illegal = false;

                    ushort src_value = 0;

                    switch (src)
                    {
                        case "d0":
                            src_value = Register[(int)CPURegister.D0];
                            break;
                        case "d1":
                            src_value = Register[(int)CPURegister.D1];
                            break;
                        case "d2":
                            src_value = Register[(int)CPURegister.D2];
                            break;
                        case "d3":
                            src_value = Register[(int)CPURegister.D3];
                            break;
                        case "d4":
                            src_value = Register[(int)CPURegister.D4];
                            break;
                        case "d5":
                            src_value = Register[(int)CPURegister.D5];
                            break;
                        case "d6":
                            src_value = Register[(int)CPURegister.D6];
                            break;
                        case "d7":
                            src_value = Register[(int)CPURegister.D7];
                            break;
                        case "mar":
                            src_value = Register[(int)CPURegister.MAR];
                            break;
                        case "mdr":
                            src_value = Register[(int)CPURegister.MDR];
                            break;
                        case "alux":
                            src_value = Register[(int)CPURegister.ALUx];
                            break;
                        case "aluy":
                            src_value = Register[(int)CPURegister.ALUy];
                            break;
                        case "alur":
                            src_value = Register[(int)CPURegister.ALUr];
                            break;
                        case "usp":
                            src_value = Register[(int)CPURegister.USP];
                            break;
                        case "ssp":
                            src_value = Register[(int)CPURegister.SSP];
                            break;
                        case "ir(operand)":
                            src_value = Register[(int)CPURegister.IRoperand];
                            break;
                        case "pc":
                            src_value = Register[(int)CPURegister.PC];
                            break;
                        case "psr":
                            src_value = Register[(int)CPURegister.PSR];
                            break;
                        case "ir(cmd)":
                            src_value = Register[(int)CPURegister.IRcmd];
                            break;
                        case "ir(r1)":
                            // first get r1 by masking IR with 000F
                            r = Register[(int)CPURegister.IRcmd] & 0x000F;
                            // then lookup the r'th value in the list of registers indexed this way
                            src_value = Register[r];

                            //Console.WriteLine("ir(r1) = " + r);

                            break;
                        case "ir(r0)":
                            // first get r0 by masking IR with 00F0 and diving by 16
                            r = Register[(int)CPURegister.IRcmd] & 0x00F0;
                            r = r / 16;
                            // then lookup the r'th value in the list of registers indexed this way
                            src_value = Register[r];

                            //Console.WriteLine("ir(r0) = " + r);

                            break;
                        case "jumpers(intbase)":
                            src_value = interruptBase;
                            break;
                        case "psr(intvec)":
                            src_value = (ushort)Vector;
                            break;
                        default:
                            src_illegal = true;
                            break;
                    }

                    if (src_illegal)
                    {
                        // do what?
                        messageList.AddError("illegal src [" + src + "]");
                        //Console.WriteLine("illegal src [" + src + "]");
                    }
                    else
                    {
                        switch (dest)
                        {
                            case "d0":
                                Register[(int)CPURegister.D0] = src_value;
                                break;
                            case "d1":
                                Register[(int)CPURegister.D1] = src_value;
                                break;
                            case "d2":
                                Register[(int)CPURegister.D2] = src_value;
                                break;
                            case "d3":
                                Register[(int)CPURegister.D3] = src_value;
                                break;
                            case "d4":
                                Register[(int)CPURegister.D4] = src_value;
                                break;
                            case "d5":
                                Register[(int)CPURegister.D5] = src_value;
                                break;
                            case "d6":
                                Register[(int)CPURegister.D6] = src_value;
                                break;
                            case "d7":
                                Register[(int)CPURegister.D7] = src_value;
                                break;
                            case "mar":
                                Register[(int)CPURegister.MAR] = src_value;
                                break;
                            case "mdr":
                                Register[(int)CPURegister.MDR] = src_value;
                                break;
                            case "alux":
                                Register[(int)CPURegister.ALUx] = src_value;
                                break;
                            case "aluy":
                                Register[(int)CPURegister.ALUy] = src_value;
                                break;
                            case "alur":
                                Register[(int)CPURegister.ALUr] = src_value;
                                break;
                            case "usp":
                                Register[(int)CPURegister.USP] = src_value;
                                break;
                            case "ssp":
                                Register[(int)CPURegister.SSP] = src_value;
                                break;
                            case "ir(operand)":
                                Register[(int)CPURegister.IRoperand] = src_value;
                                break;
                            case "pc":
                                Register[(int)CPURegister.PC] = src_value;
                                break;
                            case "psr":
                                Register[(int)CPURegister.PSR] = src_value;
                                break;
                            case "ir(cmd)":
                                Register[(int)CPURegister.IRcmd] = src_value;
                                break;
                            case "ir(r1)":
                                // first get r1 by masking IR with 000F
                                r = Register[(int)CPURegister.IRcmd] & 0x000F;
                                // then update the r'th value in the list of registers indexed this way
                                Register[r] = src_value;
                                break;
                            case "ir(r0)":
                                // first get r0 by masking IR with 00F0 and diving by 16
                                r = Register[(int)CPURegister.IRcmd] & 0x00F0;
                                r = r / 16;
                                // then lookup the r'th value in the list of registers indexed this way
                                Register[r] = src_value;
                                break;
                            case "psr(intvec)":
                                Vector = src_value;
                                break;
                            default:
                                dest_illegal = true;
                                break;
                        }

                        if (dest_illegal)
                        {
                            // do what?
                            messageList.AddError("illegal dest [" + dest + "]");
                            //Console.WriteLine("illegal dest [" + dest + "]");
                        }
                    }


                    break;

            } // end of the switch statement
            return (RetVal);
        }

        public void RunMicroInstructionSet(ushort opcode, ref Memory Memory)
        {
            ArrayList Opcodes = new ArrayList();
            int returncode = 0;

            //string error = "";

            Opcodes = firmware.getMicroInstructions(opcode); 
            if (Opcodes.Count > 0)
            {
               foreach ( string microinstruction in Opcodes )
               {
                    returncode = RunSingleMicroInstruction(microinstruction, ref Memory);

                    // should we check I/O in here?

                    if (returncode == 1) return;
                }
            }
            //else
            //{
            //    if (opcode == CPU.FETCH)
            //    {
            //        //error = "No microcodes defined for fetch, have you loaded an instruction set?";
            //        //MessageStore.AddMessage(messagesType.Error, 0, error);
            //        //messageList.AddError(error);
            //        running = false;
            //    }
            //    else
            //    {
            //        //error = "No microcodes defined for opcode 0x" +
            //        //    opcode.ToString("X").PadLeft(2, '0').ToLower() + ", have you loaded an instruction set?";
            //        //MessageStore.AddMessage(messagesType.Error, 0, error);
            //        //messageList.AddError(error);
            //    }
            //}
            return;
        }

        public void RunInterruptCycle(ref Memory Memory)
        {
            ushort opcode = CPU.INTERRUPT;
            RunMicroInstructionSet(opcode, ref Memory);
            return;
        }

        public void RunFetchCycle(ref Memory Memory)
        {
            ushort opcode = CPU.FETCH;
            RunMicroInstructionSet(opcode, ref Memory);
            return;
        }

        public void RunExecuteCycle(ref Memory Memory)
        {
            ushort opcode = (ushort)(Register[(int)CPURegister.CU] % 256);
            RunMicroInstructionSet(opcode, ref Memory);
            return;
        }

        public void RunFetchExecuteCycle(ref Memory Memory)
        {
            ushort vector = 0;

            RunFetchCycle(ref Memory);
            RunExecuteCycle(ref Memory);

            if ((GetFlag(PSRflag.I) == 1) && (Vector == 7)) {
                // we have an NMI
                RunInterruptCycle(ref Memory);
            }
            else if ((GetFlag(PSRflag.I) == 1) && (GetFlag(PSRflag.E) == 1))
            {
                // if interrupts enabled and interrupt requested then fire interrupt handler
                RunInterruptCycle(ref Memory);
            }
            else if (GetFlag(PSRflag.E) == 1)
            {
                // check the timer to see if an interrupt has been fired
                if (Memory.ioChip.Installed)
                {
                    if (Memory.clockChip.InterruptRaised(ref vector)) {
                        Vector = vector;
                        SetFlag(PSRflag.I, 1);
                    }
                }
            }
            return;
        }

        public string[] GetInstructionSetDetails()
        {
            // maximum length for stringarray = opcodes + leading line
            string[] stringarray = new string[257];
            int index = 0;
            bool foundInstruction = false;
            string s = "";
            string error = "";

            s = "Loaded microcodes";
            stringarray[index] = s;
            index++;

            for (ushort i = 0; i < 256; i++)
            {
                if (firmware.getMnemonic(i).Equals("", StringComparison.CurrentCultureIgnoreCase))
                {
                    // do nothing
                }
                else
                {
                    foundInstruction = true;
                    s = i.ToString("X").PadLeft(2, '0').ToLower() + " "
                        + firmware.getMnemonic(i).PadRight(20, ' ');
                    s = s + firmware.getDescription(i);
                    stringarray[index] = s;
                    index++;
                }
            }
            if (foundInstruction == false)
            {
                error = "No instruction set loaded";
                //Console.WriteLine(error);

                messageList.AddError(error);
                //MessageStore.AddMessage(messagesType.Information, 0, error);
            }

            // build return string
            string[] returnstring = new string[index];
            for (int i = 0; i < index; i++)
            {
                returnstring[i] = stringarray[i];
            }

            return (returnstring);
        }

        public string[] OpcodeDetails(ushort opcode)
        {
            int index = 0;
            int size = firmware.getMicroInstructions(opcode).Count + 3;
            string[] s = new string[size];
            // display the instruction
            s[index] = "Opcode " + opcode.ToString("X").PadLeft(2, '0').ToLower();
            index++;
            s[index] = "Mnemonic \"" + firmware.getMnemonic(opcode) + "\"";
            index++;
            s[index] = "Description \"" + firmware.getDescription(opcode) + "\"";
            index++;
            foreach (string m in firmware.getMicroInstructions(opcode))
            {
                s[index] = m;
                index++;
            }

            return (s);
        }

        public string[] OpcodeRawDetails(ushort opcode)
        {
            int index = 0;
            int size = firmware.getRawMicroInstructions(opcode).Count + 3;
            string[] s = new string[size];
            // display the instruction
            s[index] = "Opcode " + opcode.ToString("X").PadLeft(2, '0').ToLower();
            index++;
            s[index] = "Mnemonic \"" + firmware.getMnemonic(opcode) + "\"";
            index++;
            s[index] = "Description \"" + firmware.getDescription(opcode) + "\"";
            index++;
            foreach (string m in firmware.getRawMicroInstructions(opcode))
            {
                s[index] = m;
                index++;
            }

            return (s);
        }

        public ArrayList OpcodeMicroInstructions(ushort i)
        {
            return (firmware.getMicroInstructions(i));
        }

        public ushort MicromemoryLimit
        {
            get
            {
                return (firmware.Limit);
            }
        }

        public void Wipe()
        {
            // init opcodes
            firmware.clearAllOpcodes();
        }

        public void HardReset()
        {
            for (int i = 0; i < REGISTERSIZE; i++)
            {
                Register[i] = 0;
            }
            machinecycles = 0;
            running = false;
            return;
        }

        public void SoftReset()
        {
            running = false;
            return;
        }

        public void SetRegister(CPURegister r, ushort value)
        {
            string error = "";

            if ((r == CPURegister.CU) && (value > 255))
            {
                error = "Number format invalid";
                //Console.WriteLine(error);
                messageList.AddError(error);

                //MessageStore.AddMessage(messagesType.Error, 0, error);
            }
            else
            {
                Register[(int)r] = value;
            }
            return;
        }

        public ushort GetRegister(CPURegister r)
        {
            return (Register[(int)r]);
        }

        public string GetRegister(CPURegister r, Base b)
        {
            string s = "";
            if (r == CPURegister.CU) {
                s = supportMethods.FormatUshort((ushort)Register[(int)r], b, 2);
            }
            else {
                s = supportMethods.FormatUshort((ushort)Register[(int)r], b);
            }
            return (s);
        }

        public ushort GetFlag(PSRflag f)
        {
            ushort flag = 0;
            ushort mask = 0;
            switch (f)
            {
                case PSRflag.V:
                    mask = maskV;
                    break;
                case PSRflag.N:
                    mask = maskN;
                    break;
                case PSRflag.Z:
                    mask = maskZ;
                    break;
                case PSRflag.C:
                    mask = maskC;
                    break;
                case PSRflag.I:
                    mask = maskI;
                    break;
                case PSRflag.E:
                    mask = maskE;
                    break;
                default:
                    break;
            }
            if ((Register[(int)CPURegister.PSR] & mask) == 0)
                flag = 0;
            else
                flag = 1;
            return (flag);
        }

        public void SetFlag(PSRflag f, ushort value)
        {
            string error = "";
            ushort mask = 0;
            if ((value != 0) && (value != 1))
            {
                error = "Number format invalid";
                //Console.WriteLine(error);
                messageList.AddError(error);

                //MessageStore.AddMessage(messagesType.Error, 0, error);
            }
            else
            {
                switch (f)
                {
                    case PSRflag.V:
                        mask = maskV;
                        break;
                    case PSRflag.N:
                        mask = maskN;
                        break;
                    case PSRflag.Z:
                        mask = maskZ;
                        break;
                    case PSRflag.C:
                        mask = maskC;
                        break;
                    case PSRflag.I:
                        mask = maskI;
                        break;
                    case PSRflag.E:
                        mask = maskE;
                        break;
                    case PSRflag.S:
                        mask = maskS;
                        break;
                    default:
                        break;
                }
                if (value == 0)
                    Register[(int)CPURegister.PSR] 
                        = (ushort)(Register[(int)CPURegister.PSR] & (~mask));
                else
                    Register[(int)CPURegister.PSR] 
                        = (ushort)(Register[(int)CPURegister.PSR] | mask);
            }
            return;
        }

        public bool IsRunning
        {
            get
            {
                return (running);
            }
        }

        public void Start()
        {
            running = true;
        }

        public ushort Vector
        {
            set
            {
                string error = "";
                ushort result = value;
                int tempresult = 0;
                int temp = 0;

                if ((result >= 0) && (result <= 7))
                {
                    // update vector v3
                    if ((result / 4) == 1)
                    {
                        // top bit set
                        temp = Register[(int)CPURegister.PSR] | maskv3;
                    }
                    else
                    {
                        temp = Register[(int)CPURegister.PSR] & ~(maskv3);
                    }
                    Register[(int)CPURegister.PSR] = (ushort)temp;

                    // update vector v2
                    tempresult = result % 4;
                    result = (ushort)tempresult;
                    if ((result / 2) == 1)
                    {
                        // top bit set
                        temp = Register[(int)CPURegister.PSR] | maskv2;
                    }
                    else
                    {
                        temp = Register[(int)CPURegister.PSR] & ~(maskv2);
                    }
                    Register[(int)CPURegister.PSR] = (ushort)temp;

                    // update vector v1
                    tempresult = result % 2;
                    result = (ushort)tempresult;
                    if ((result / 1) == 1)
                    {
                        // top bit set
                        temp = Register[(int)CPURegister.PSR] | maskv1;
                    }
                    else
                    {
                        temp = Register[(int)CPURegister.PSR] & ~(maskv1);
                    }
                    Register[(int)CPURegister.PSR] = (ushort)temp;
                }
                else
                {
                    error = "Number format invalid";
                    //Console.WriteLine(error);
                    messageList.AddError(error);

                    //MessageStore.AddMessage(messagesType.Error, 0, error);
                }
            }

            get
            {
                ushort vector = 0;

                ushort v1 = 0;
                ushort v2 = 0;
                ushort v3 = 0;

                if ((Register[(int)CPURegister.PSR] & maskv3) == maskv3)
                {
                    v3 = 1;
                }

                if ((Register[(int)CPURegister.PSR] & maskv2) == maskv2)
                {
                    v2 = 1;
                }

                if ((Register[(int)CPURegister.PSR] & maskv1) == maskv1)
                {
                    v1 = 1;
                }

                vector = (ushort)((4 * v3) + (2 * v2) + v1);
                return (vector);
            }
        }

        public string Version
        {
            get
            {
                return (CPU_VER);
            }
        }

        public ulong MachineCycles
        {
            get
            {
                return (machinecycles);
            }
            set
            {
                machinecycles = value;
            }
        }

        public ushort InterruptBase
        {
            get
            {
                return (interruptBase);
            }
            set
            {
                interruptBase = value;
            }
        }

        private void UpdateFlagZ()
        {
            if (Register[(int)CPURegister.ALUr] == 0) 
                SetFlag(PSRflag.Z, 1); 
            else 
                SetFlag(PSRflag.Z, 0);
            return;
        }

        private void UpdateFlagN()
        {
            if ((Register[(int)CPURegister.ALUr] & 0x8000) == 0x8000)
                SetFlag(PSRflag.N, 1);
            else
                SetFlag(PSRflag.N, 0);
            return;
        }        

        private void UpdateFlagC(int i)
        {
            if ((i & 0x00010000) == 0x00010000)
                SetFlag(PSRflag.C, 1);
            else
                SetFlag(PSRflag.C, 0);
            return;
        }        

    }
   
}