//Copyright (c) 2009, Mark S. Burrell
//All rights reserved.

//Redistribution and use in source and binary forms, with or without modification, are 
//permitted provided that the following conditions are met:

//    * Redistributions of source code must retain the above copyright notice, this list 
//of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this 
//list of conditions and the following disclaimer in the documentation and/or other 
//materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
//EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
//OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
//SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
//OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;


namespace Sedici.Hardware
{
    public class Firmware
    {

        // positions in the opcode arrays
        private const ushort FETCH = 256;
        private const ushort INTERRUPT = 257;

        private string comment = null;

        // opcode class for use by the instruction set
        private class opcode
        {

            private ArrayList rawmicroinstructions; // cludge to cope with ALL info from an opcode definition including comments etc.
            private ArrayList microinstructions;
            private string mnemonic;
            private string description;

            public opcode()
            {
                rawmicroinstructions = new ArrayList();
                microinstructions = new ArrayList();
                Clear();
            }

            public string Description
            {
                get
                {
                    return (description);
                }
                set
                {
                    description = value;
                }
            }

            public string Mnemonic
            {
                get
                {
                    return (mnemonic);
                }
                set
                {
                    mnemonic = value;
                }
            }

            public ArrayList MicroInstructions
            {
                get
                {
                    return (microinstructions);
                }
            }

            public ArrayList RawMicroInstructions
            {
                get
                {
                    return (rawmicroinstructions);
                }
            }

            public void AddMicroInstruction(string s)
            {
                microinstructions.Add(s);
                return;
            }

            public void AddRawMicroInstruction(string s)
            {
                rawmicroinstructions.Add(s);
                return;
            }

            public void Clear()
            {
                rawmicroinstructions.Clear();
                microinstructions.Clear();
                mnemonic = "";
                description = "";
            }
        }

        // base initial settings

        // storage for opcodes
        private const ushort LIMIT = 258;
        private opcode[] Opcodes = new opcode[LIMIT];

        public Firmware()
        {
            for (int i = 0; i < LIMIT; i++ )
            {
                Opcodes[i] = new opcode();
            }
            clearAllOpcodes();
            return;
        }

        public string Comment
        {
            get
            {
                return (comment);
            }
            set
            {
                if (comment == null)
                {
                    comment = value;
                }
                else
                {
                    comment = comment + Environment.NewLine + value;
                }
            }
        }

        public void clearComment()
        {
            comment = null;
        }

        public ushort Limit
        {
            get
            {
                return (LIMIT);
            }
        }

        public void setSingleMicroInstruction(ushort i, string s)
        {
            // if index is invalid we do nothing
            if (i < LIMIT)
            {
                //opcode[i] = opcode[i]  + "\0" + s;
                Opcodes[i].AddMicroInstruction(s);
            }
        }

        public void setSingleRawMicroInstruction(ushort i, string s)
        {
            // if index is invalid we do nothing
            if (i < LIMIT)
            {
                //opcode[i] = opcode[i]  + "\0" + s;
                Opcodes[i].AddRawMicroInstruction(s);
            }
        }

        public ArrayList getMicroInstructions(ushort i)
        {
            if (i < LIMIT)
            {
                return (Opcodes[i].MicroInstructions);
            }
            else
            {
                ArrayList newarray = new ArrayList();
                return (newarray);
            }
        }

        public ArrayList getRawMicroInstructions(ushort i)
        {
            if (i < LIMIT)
            {
                return (Opcodes[i].RawMicroInstructions);
            }
            else
            {
                ArrayList newarray = new ArrayList();
                return (newarray);
            }
        }

        public string validMicroInstruction(string s)
        {
            // check to see the microinstruction is valid, otherwise return 'illegal'

            string retstring = "";

            // we do this because all micro-instructions can be written in the
            // instruction set case-insensitive
            string lower_s = s.ToLower();

            switch (lower_s)
            {

                // flags
                case "if(psr(c)==1)":
                case "if(psr(c)==0)":
                case "if(psr(n)==1)":
                case "if(psr(n)==0)":
                case "if(psr(z)==1)":
                case "if(psr(z)==0)":
                case "if(psr(v)==1)":
                case "if(psr(v)==0)":

                // ALU microinstructions
                case "alur=[alux]+[aluy]":
                case "alur=[alux]+[aluy]+[psr(c)]":
                case "alur=[alux]-[aluy]":
                case "alur=[alux]<<1":
                case "alur=[alux]>>1":
                case "alur=[alux]<<<1":
                case "alur=[alux]>>>1":
                case "alur=[alux]=>>1":
                case "alur=[alux]<<=1":
                case "alur=[alux]&[aluy]":
                case "alur=[alux]|[aluy]":
                case "alur=~[alux]":
                case "alur=[alux]+1":
                case "alur=[alux]-1":

                // general control
                case "nop":
                case "halt":
                
                // interrupt flags
                case "psr(i)=1":
                case "psr(i)=0":
                case "psr(e)=1":
                case "psr(e)=0":
                
                // specific data movements
                case "cu<-[ir(opcode)]":
                case "m[mar]<-[mdr]":
                case "mdr<-[m[mar]]":


                    retstring = s;
                    break;
                default:
                    // it's now either a valid data movement or an illegal instruction

                    // valid data movement is in the form DEST<-[SOURCE]
                    // where SOURCE/DEST can be from the following:
                    //    D0-D7, MAR, MDR, ALUx, ALUy, ALUr, USP, SSP
                    //    PC, PSR, IR, INC, IR(r1), IR(r0),
                    //    Jumpers(IntBase), PSR(IntVec) 

                    // The first thing to do is make sure we have something in the right format
                    // and extract SOURCE and DEST

                    string[] move = { "<-" };
                    string[] leftbracket = { "[" };
                    string[] rightbracket = { "]" };

                    string[] parts = { };

                    string src = "";
                    string dest = "";

                    parts = lower_s.Split(move, StringSplitOptions.None);

                    if (parts.Length == 2) {
                        // we have the right number of params
                        dest = parts[0];

                        // now let's see if we can remove the brackets of the second param
                        parts = parts[1].Split(leftbracket, StringSplitOptions.None);

                        // we should have 2 parts again
                        if (parts.Length == 2)
                        {
                            parts = parts[1].Split(rightbracket, StringSplitOptions.None);

                            // again 2 parts left hopefully
                            if (parts.Length == 2)
                            {
                                // first one should be the source
                                src = parts[0];
                            }
                        }
                    }

                    // Now check if the SOURCE|DEST pair are both in the 'approved' list
                    bool src_illegal = false;
                    bool dest_illegal = false;

                    switch (src)
                    {
                        case "d0":
                        case "d1":
                        case "d2":
                        case "d3":
                        case "d4":
                        case "d5":
                        case "d6":
                        case "d7":
                        case "mar":
                        case "mdr":
                        case "alux":
                        case "aluy":
                        case "alur":
                        case "usp":
                        case "ssp":
                        case "pc":
                        case "psr":
                        case "ir(cmd)":
                        case "inc":
                        case "ir(operand)":
                        case "ir(r0)":
                        case "ir(r1)":
                        case "jumpers(intbase)":
                        case "psr(intvec)":
                            break;
                        default:
                            src_illegal = true;
                            break;
                    }

                    switch (dest)
                    {
                        case "d0":
                        case "d1":
                        case "d2":
                        case "d3":
                        case "d4":
                        case "d5":
                        case "d6":
                        case "d7":
                        case "mar":
                        case "mdr":
                        case "alux":
                        case "aluy":
                        case "alur":
                        case "usp":
                        case "ssp":
                        case "pc":
                        case "psr":
                        case "ir(cmd)":
                        case "inc":
                        case "ir(r0)":
                        case "ir(r1)":
                        case "ir(operand)":
                        case "jumpers(intbase)":
                        case "psr(intvec)":
                            break;
                        default:
                            dest_illegal = true;
                            break;
                    }

                    if (dest_illegal || src_illegal)
                    {
                        retstring = "illegal";
                    }
                    else
                    {
                        // one last check
                        if (src == dest)
                        {
                            retstring = "illegal";
                        }
                        else
                        {
                            retstring = s;
                        }
                    }
                    break;
            } // end of the switch statement
                        
            return(retstring);
        }

        public void clearAllOpcodes()
        {
            for (ushort i = 0; i < LIMIT; i++)
            {
                clearOpcode(i);
            }
            comment = null;
            return;
        }

        public void clearOpcode(ushort i)
        {
            if (i < LIMIT)
            {
                Opcodes[i].Clear();
            }
            return;
        }

        opcode getOpcode(ushort i)
        {
            if (i < LIMIT)
            {
                return (Opcodes[i]);
            }
            else
            {
                opcode newcode = new opcode();
                return (newcode);
            }
        }

        public string getMnemonic(ushort i)
        {
            // if index is invalid we return an empty string
            string s = "";
            if (i < LIMIT)
            {
                s = Opcodes[i].Mnemonic;
            }
            return (s);
        }

        public void setMnemonic(ushort i, string s)
        {
            // if index is invalid then we do nothing
            if (i < LIMIT)
            {
                Opcodes[i].Mnemonic = s;
            }
            return;
        }

        public string getDescription(ushort i)
        {
            // if index is invalid we return an empty string
            string s = "";
            if (i < LIMIT)
            {
                s = Opcodes[i].Description;
            }
            return (s);
        }

        public void setDescription(ushort i, string s)
        {
            // if index is invalid then we do nothing
            if (i < LIMIT)
            {
                Opcodes[i].Description = s;
            }
            return;
        }

        private ArrayList readTokens(string a) {


            // here's the rules:
            // -- spaces are allowed inside matching ' chars
            // -- \' is allowed inside matching ' chars
            // -- *  is allowed inside matching ' chars
            // otherwise a space divides tokens and is not stored.

            ArrayList tokens = new ArrayList();

            String nextToken = "";

            char last = '0';
            char current;

            Boolean inText = false;
            Boolean comment = false;

            for (int i = 0; i < a.Length; i++)
            {
                current = a[i];

                if ((current == ' ') && (inText == false) && (comment == false) )
                {
                    // space char not stored, create new token
                    if (!(nextToken.ToString() == ""))
                    {
                        tokens.Add(nextToken);
                        nextToken = "";
                    }
                }
                else if ((current == ' ') && (inText == true) && (comment == false))
                {
                    nextToken = nextToken + current;
                }
                else if ((current == '*') && (inText == false) && (comment == false))
                {
                    nextToken = nextToken + current;
                    comment = true;
                    // it's a comment to the end of the line
                }
                else if ((current == '*') && (inText == true) && (comment == false))
                {
                    nextToken = nextToken + current;
                }
                else if (current == '\"')
                {
                    // found a "
                    if (last == '\\')
                    {
                        // escaped '
                        nextToken = nextToken + current;
                    }
                    else
                    {
                        if (inText == false)
                        {
                            inText = true;
                        }
                        else
                        {
                            inText = false;
                        }
                        nextToken = nextToken + current;
                    }
                }
                else
                {
                    if (comment == false)
                    {
                        //tokens.Add(current);
                        nextToken = nextToken + current;
                    }
                    else
                    {
                        nextToken = nextToken + current;
                    }
                }

                last = current;
            }

            // add any non-blank token left over
            if (!(nextToken.ToString() == ""))
            {
                tokens.Add(nextToken);
            }


            return (tokens);
        }

        public void LoadFirmware(string filename)
        {
            string contents = "";
            string error = "";
            string tempstring = "";
            string[] lines = { };
            string[] line = { };
            string[] separator = { Environment.NewLine };
            string[] spaceseparator = { " " };
            string[] textseparator = { "\"" };
            string[] colonseparator = { ":" };

            ArrayList tokens;

            ushort result = 0;
            int lineNumber = 0;

            bool opcodeDefined = false;
            ushort currentOpcode = 0;

            if (File.Exists(filename))
            {
                // read the file
                contents = File.ReadAllText(filename);

                lines = contents.Split(separator, StringSplitOptions.None);

                for (int count = 0; count < lines.Length; count++)
                {
                    tokens = new ArrayList();

                    string currentLine = lines[count]; 

                    // turn the line into tokens
                    tokens = readTokens(currentLine);

                    // process the tokens for the line
                    if (tokens.Count > 0) {

                        if (tokens[0].ToString().Equals("opcode", StringComparison.CurrentCultureIgnoreCase))
                        {
                            if (tokens.Count > 1)
                            {
                                if (tokens[1].ToString().Equals("fetch", StringComparison.CurrentCultureIgnoreCase))
                                {
                                    opcodeDefined = true;
                                    currentOpcode = FETCH;
                                    clearOpcode(currentOpcode);
                                }
                                else if (tokens[1].ToString().Equals("interrupt", StringComparison.CurrentCultureIgnoreCase))
                                {
                                    opcodeDefined = true;
                                    currentOpcode = INTERRUPT;
                                    clearOpcode(currentOpcode);
                                }
                                else
                                {
                                    // decode opcode
                                    if (supportMethods.isNumber(tokens[1].ToString(), ref result) == 0)
                                    {
                                        lineNumber = count + 1;
                                        error = "Invalid opcode at line ";
                                        Console.WriteLine(error);

                                    }
                                    else
                                    {
                                        if (result > 255)
                                        {
                                            lineNumber = count + 1;
                                            error = "Invalid opcode at line ";
                                            Console.WriteLine(error);
                                        }
                                        else
                                        {
                                            opcodeDefined = true;
                                            currentOpcode = result;

                                            // clear out the opcode
                                            clearOpcode(currentOpcode);
                                        }
                                    }
                                }
                            }
                            else {
                                // missing opcode
                                lineNumber = count + 1;
                                error = "Missing opcode at line " + lineNumber;
                                Console.WriteLine(error);
                            }

                        }
                        else if (tokens[0].ToString().Equals("mnemonic", StringComparison.CurrentCultureIgnoreCase))
                        {
                            if (tokens.Count > 1)
                            {
                                // remove leading and training " chars
                                if (tokens[1].ToString().Length > 2)
                                {
                                    tokens[1] = tokens[1].ToString().Substring(1, (tokens[1].ToString().Length - 2));
                                    setMnemonic(currentOpcode, tokens[1].ToString());
                                }
                                else
                                {
                                    //lineNumber = count + 1;
                                    //error = "Missing mnemonic at line " + lineNumber;
                                    //Console.WriteLine(error);
                                }

                            }
                            else
                            {
                                lineNumber = count + 1;
                                error = "Missing mnemonic at line " + lineNumber;
                                Console.WriteLine(error);
                            }
                        }
                        else if (tokens[0].ToString().Equals("description", StringComparison.CurrentCultureIgnoreCase))
                        {
                            if (tokens.Count > 1)
                            {
                                // remove leading and training " chars
                                if (tokens[1].ToString().Length > 2)
                                {
                                    tokens[1] = tokens[1].ToString().Substring(1, (tokens[1].ToString().Length - 2));
                                    setDescription(currentOpcode, tokens[1].ToString());
                                }
                                else
                                {
                                    //lineNumber = count + 1;
                                    //error = "Missing description at line " + lineNumber;
                                    //Console.WriteLine(error);
                                }

                            }
                            else
                            {
                                lineNumber = count + 1;
                                error = "Missing description at line " + lineNumber;
                                Console.WriteLine(error);
                            }

                        }
                        else if (tokens[0].ToString().StartsWith("*-----"))
                        {
                            // pretty line written by firmware builder so ignore
                        }
                        else if (tokens[0].ToString().StartsWith("*"))
                        {
                            if (opcodeDefined == false)
                            {
                                //Console.WriteLine("main comment : " + currentLine);
                                Comment = tokens[0].ToString();

                            }
                            else
                            {
                                // then its a comment within firmware and the whole line should be included
                                setSingleRawMicroInstruction(currentOpcode, currentLine);
                            }
                        }
                        else
                        {
                            setSingleRawMicroInstruction(currentOpcode, currentLine);

                            tempstring = validMicroInstruction(tokens[0].ToString());

                            setSingleMicroInstruction(currentOpcode, tempstring);
                            if (tempstring == "illegal")
                            {
                                // report the error
                                lineNumber = count + 1;
                                error = "Invalid microinstruction at line " + lineNumber;
                                Console.WriteLine(error);
                            }

                        }
                    }
                }

            }
            else
            {
                error = "File [" + filename + "] does not exist in local or environment directory";
                //MessageStore.AddMessage(messagesType.Error, 0, error);


                // temp code
                Console.WriteLine(error);
            }

            return;

        }

        public void LoadFirmwareOldFormat(string filename)
        {
            // reads the new simplified firmware file

            string contents = "";
            string error = "";
            string[] tempstring = { };
            string[] lines = { };
            string[] line = { };
            string[] separator = { Environment.NewLine };
            string[] spaceseparator = { " " };
            string[] barseparator = { "!" };
            string[] textseparator = { "\"" };

            ushort result = 0;
            int lineNumber = 0;

            bool opcodeDefined = false;
            ushort currentOpcode = 0;

            comment = null;

            if (File.Exists(filename))
            {
                // read the file
                contents = File.ReadAllText(filename);

                lines = contents.Split(separator, StringSplitOptions.None);

                for (int count = 0; count < lines.Length; count++)
                {

                    //split on " " character and see if anything there other than whitespace
                    string[] wparts = lines[count].Split(spaceseparator, StringSplitOptions.RemoveEmptyEntries);
                    if (wparts.Length != 0)
                    {
                        if (lines[count].Length > 2)
                        {
                            // figure out what sort of instruction it is

                            //Console.Write("---> ");
                            //Console.WriteLine(lines[count]);
                            //split on "|" character and look at first two letters
                            string[] instructionParts = lines[count].Split(barseparator, StringSplitOptions.None);

                            if (instructionParts.Length > 1)
                            {
                                if (instructionParts[0] == "#I")
                                {
                                    //Console.WriteLine("#I   [" + lines[count] + "]");

                                    if (instructionParts.Length >= 2) {
                                        if (instructionParts[1].Equals("fetch", StringComparison.CurrentCultureIgnoreCase)) {
                                            opcodeDefined = true;
                                            currentOpcode = FETCH;
                                            clearOpcode(currentOpcode);
                                        }
                                        else if (instructionParts[1].Equals("interrupt", StringComparison.CurrentCultureIgnoreCase)) {
                                            opcodeDefined = true;
                                            currentOpcode = INTERRUPT;
                                            clearOpcode(currentOpcode);
                                        }
                                        else {
                                            if (supportMethods.isNumber(instructionParts[1].ToString(), ref result) == 0) {

                                                //count++;

                                                //Console.Write("--->");
                                                //Console.WriteLine(instructionParts[1].ToString());

                                                lineNumber = count + 1;
                                                error = "Invalid opcode at line " + lineNumber;
                                                //MessageStore.AddMessage(messagesType.Error, 0, error);

                                                // temp code
                                                Console.WriteLine(error);
                                                //return;

                                            }
                                            else {
                                                if (result > 255)
                                                {
                                                    lineNumber = count + 1;
                                                    error = "Invalid opcode at line " + lineNumber;
                                                    Console.WriteLine(error);
                                                }
                                                else
                                                {
                                                    opcodeDefined = true;
                                                    currentOpcode = result;

                                                    // clear out the opcode
                                                    clearOpcode(currentOpcode);

                                                }
                                            }
                                        }
                                    }

                                    if (instructionParts.Length >= 3)
                                    {
                                        setMnemonic(currentOpcode, instructionParts[2]);
                                    }

                                    if (instructionParts.Length >= 4) {
                                        setDescription(currentOpcode, instructionParts[3]);
                                    }
                                }
                                else if (instructionParts[0] == "#C")
                                {
                                    //Console.WriteLine("#C   [" + lines[count] + "]");
                                    
                                    if (instructionParts.Length >= 2)
                                    {
                                        Comment = instructionParts[1];
                                    }
                                    
                                }
                                else if (instructionParts[0] == "#F")
                                {
                                    if (opcodeDefined)
                                    {
                                        if (instructionParts.Length >= 2)
                                        {
                                            tempstring = instructionParts[1].Split(spaceseparator, StringSplitOptions.RemoveEmptyEntries);
                                            if (tempstring.Length > 0)
                                            {
                                                // we have the microinstruction in tempstring[0] so add it to the opcode
                                                // but check first to see if it's a comment
                                                //Console.WriteLine("[" + tempstring[0] + "]");

                                                if (tempstring[0].Substring(0,1) == "*") {
                                                    // comment
                                                    setSingleRawMicroInstruction(currentOpcode, instructionParts[1]);

                                                }
                                                else {

                                                    setSingleRawMicroInstruction(currentOpcode, instructionParts[1]);

                                                    tempstring[0] = validMicroInstruction(tempstring[0]);

                                                    setSingleMicroInstruction(currentOpcode, tempstring[0]);
                                                    if (tempstring[0] == "illegal")
                                                    {
                                                        // report the error
                                                        lineNumber = count + 1;
                                                        error = "Invalid microinstruction at line " + lineNumber;
                                                        Console.WriteLine(error);
                                                    }
                                                }

                                            }
                                            else
                                            {
                                                // blank
                                                setSingleRawMicroInstruction(currentOpcode, "");
                                            }


                                        }
                                        else
                                        {
                                            // deformed instruction?
                                            lineNumber = count + 1;
                                            error = "Invalid firmware at line " + lineNumber;
                                            Console.WriteLine(error);
                                        }

                                    }
                                    else
                                    {
                                        lineNumber = count + 1;
                                        error = "Using microcode before opcode defined at line " + lineNumber;
                                        Console.WriteLine(error);
                                    }
                                    //Console.WriteLine("#F   [" + lines[count] + "]");
                                }
                                else
                                {
                                    // deformed instruction?
                                    lineNumber = count + 1;
                                    error = "Invalid firmware at line " + lineNumber;
                                    Console.WriteLine(error);
                                }
                            }
                            else
                            {
                                // deformed instruction?
                                lineNumber = count + 1;
                                error = "Invalid firmware at line " + lineNumber;
                                Console.WriteLine(error);
                            }

                        }
                        else
                        {
                            // deformed instruction?
                            lineNumber = count + 1;
                            error = "Invalid firmware at line " + lineNumber;
                            Console.WriteLine(error);
                        }
                    }
                }
            }
            else
            {
                error = filename + " - failed to open";
                Console.WriteLine(error);
            }

            return;
        }


        //public ushort interruptBase
        //{
        //    get
        //    {
        //        return (interruptbase);
        //    }
        //    set
        //    {
        //        interruptbase = value;
        //    }
        //}

    } // end class
}
